"""
This module is the computational part of the geometrical module of ToFu
"""

# Built-in
import warnings

# Common
import numpy as np
import scipy.interpolate as scpinterp
import scipy.integrate as scpintg
from inspect import signature as insp

# ToFu-specific
try:
    import tofu.geom._def as _def
    import tofu.geom._GG as _GG
except Exception:
    from . import _def as _def
    from . import _GG as _GG


###############################################################################
#                            Default parameters
###############################################################################


_SAMPLE_RES = {
    'edge': 0.02,
    'cross': 0.1,
    'surface': 0.1,
    'volume': 0.1,
}
_SAMPLE_RESMODE = {
    'edge': 'abs',
    'cross': 'abs',
    'surface': 'abs',
    'volume': 'abs',
}

###############################################################################
#                            Ves functions
###############################################################################


# ==============================================================================
# = Ves sub-functions
# ==============================================================================
def _Struct_set_Poly(
    Poly, pos=None, extent=None, arrayorder="C", Type="Tor", Clock=False
):
    """ Compute geometrical attributes of a Struct object """

    # Make Poly closed, counter-clockwise, with '(cc,N)' layout and arrayorder
    try:
        Poly = _GG.format_poly(Poly, order="C", Clock=False, close=True,
                               Test=True)
    except Exception as excp:
        print(excp)
    assert Poly.shape[0] == 2, "Arg Poly must be a 2D polygon !"
    fPfmt = np.ascontiguousarray if arrayorder == "C" else np.asfortranarray

    # Get all remarkable points and moments
    NP = Poly.shape[1] - 1
    P1Max = Poly[:, np.argmax(Poly[0, :])]
    P1Min = Poly[:, np.argmin(Poly[0, :])]
    P2Max = Poly[:, np.argmax(Poly[1, :])]
    P2Min = Poly[:, np.argmin(Poly[1, :])]
    BaryP = np.sum(Poly[:, :-1], axis=1, keepdims=False) / (Poly.shape[1] - 1)
    BaryL = np.array(
        [(P1Max[0] + P1Min[0]) / 2.0, (P2Max[1] + P2Min[1]) / 2.0]
    )
    BaryS, Surf = _GG.poly_area_and_barycenter(Poly, NP)

    # Get lim-related indicators
    noccur = int(pos.size)
    Multi = noccur > 1

    # Get Tor-related quantities
    if Type.lower() == "lin":
        Vol, BaryV = None, None
    else:
        Vol, BaryV = _GG.Poly_VolAngTor(Poly)
        if Vol <= 0.0:
            msg = ("Pb. with volume computation for Struct of type 'Tor' !\n"
                   + "\t- Vol = {}\n".format(Vol)
                   + "\t- Poly = {}\n\n".format(str(Poly))
                   + "  => Probably corrupted polygon\n"
                   + "  => Please check polygon is not self-intersecting")
            raise Exception(msg)

    # Compute the non-normalized vector of each side of the Poly
    Vect = np.diff(Poly, n=1, axis=1)
    Vect = fPfmt(Vect)

    # Compute the normalised vectors directed inwards
    Vin = np.array([Vect[1, :], -Vect[0, :]])
    Vin = -Vin  # Poly is Counter Clock-wise as defined above
    Vin = Vin / np.hypot(Vin[0, :], Vin[1, :])[np.newaxis, :]
    Vin = fPfmt(Vin)

    poly = _GG.format_poly(
        Poly,
        order=arrayorder,
        Clock=Clock,
        close=False,
        Test=True,
    )

    # Get bounding circle
    circC = BaryS
    r = np.sqrt(np.sum((poly - circC[:, np.newaxis]) ** 2, axis=0))
    circr = np.max(r)

    dout = {
        "Poly": poly,
        "pos": pos,
        "extent": extent,
        "noccur": noccur,
        "Multi": Multi,
        "nP": NP,
        "P1Max": P1Max,
        "P1Min": P1Min,
        "P2Max": P2Max,
        "P2Min": P2Min,
        "BaryP": BaryP,
        "BaryL": BaryL,
        "BaryS": BaryS,
        "BaryV": BaryV,
        "Surf": Surf,
        "VolAng": Vol,
        "Vect": Vect,
        "VIn": Vin,
        "circ-C": circC,
        "circ-r": circr,
        "Clock": Clock,
    }
    return dout


def _Ves_get_InsideConvexPoly(
    Poly,
    P2Min,
    P2Max,
    BaryS,
    RelOff=_def.TorRelOff,
    ZLim="Def",
    Spline=True,
    Splprms=_def.TorSplprms,
    NP=_def.TorInsideNP,
    Plot=False,
    Test=True,
):
    if Test:
        assert type(RelOff) is float, "Arg RelOff must be a float"
        assert (
            ZLim is None or ZLim == "Def" or type(ZLim) in [tuple, list]
        ), "Arg ZLim must be a tuple (ZlimMin, ZLimMax)"
        assert type(Spline) is bool, "Arg Spline must be a bool !"
    if ZLim is not None:
        if ZLim == "Def":
            ZLim = (
                P2Min[1] + 0.1 * (P2Max[1] - P2Min[1]),
                P2Max[1] - 0.05 * (P2Max[1] - P2Min[1]),
            )
        indZLim = (Poly[1, :] < ZLim[0]) | (Poly[1, :] > ZLim[1])
        if Poly.shape[1] - indZLim.sum() < 10:
            msg = "Poly seems to be Convex and simple enough !"
            msg += "\n  Poly.shape[1] - indZLim.sum() < 10"
            warnings.warn(msg)
            return Poly
        Poly = np.delete(Poly, indZLim.nonzero()[0], axis=1)
    if np.all(Poly[:, 0] == Poly[:, -1]):
        Poly = Poly[:, :-1]
    Np = Poly.shape[1]
    if Spline:
        BarySbis = np.tile(BaryS, (Np, 1)).T
        Ptemp = (1.0 - RelOff) * (Poly - BarySbis)
        # Poly = BarySbis + Ptemp
        Ang = np.arctan2(Ptemp[1, :], Ptemp[0, :])
        Ang, ind = np.unique(Ang, return_index=True)
        Ptemp = Ptemp[:, ind]
        # spline parameters
        ww = Splprms[0] * np.ones((Np + 1,))
        ss = Splprms[1] * (Np + 1)  # smoothness parameter
        kk = Splprms[2]  # spline order
        nest = int(
            (Np + 1) / 2.0
        )  # estimate of number of knots needed (-1 = maximal)
        # Find the knot points

        # TODO @DV : we can probably get rid of this
        # tckp,uu = scpinterp.splprep([np.append(Ptemp[0,:],Ptemp[0,0]),
        # np.append(Ptemp[1,:],Ptemp[1,0]),np.append(Ang,Ang[0]+2.*np.pi)],
        # w=ww, s=ss, k=kk, nest=nest)
        tckp, uu = scpinterp.splprep(
            [
                np.append(Ptemp[0, :], Ptemp[0, 0]),
                np.append(Ptemp[1, :], Ptemp[1, 0]),
            ],
            u=np.append(Ang, Ang[0] + 2.0 * np.pi),
            w=ww,
            s=ss,
            k=kk,
            nest=nest,
            full_output=0,
        )
        xnew, ynew = scpinterp.splev(np.linspace(-np.pi, np.pi, NP), tckp)
        Poly = np.array([xnew + BaryS[0], ynew + BaryS[1]])
        Poly = np.concatenate((Poly, Poly[:, 0:1]), axis=1)
    if Plot:
        import matplotlib.pyplot as plt

        f = plt.figure(facecolor="w", figsize=(8, 10))
        ax = f.add_axes([0.1, 0.1, 0.8, 0.8])
        ax.plot(Poly[0, :], Poly[1, :], "-k", Poly[0, :], Poly[1, :], "-r")
        ax.set_aspect(aspect="equal", adjustable="datalim"), ax.set_xlabel(
            r"R (m)"
        ), ax.set_ylabel(r"Z (m)")
        f.canvas.draw()
    return Poly


# ==============================================================================
# = Ves sampling functions
# ==============================================================================


def _Ves_get_sample_checkinputs(
    res=None,
    domain=None,
    resMode=None,
    ind=None,
    which='volume'
):
    """ Check inputs for all sampling routines """

    # res
    dres = {'edge': 1, 'cross': 2, 'surface': 2, 'volume': 3}
    if res is None:
        res = _SAMPLE_RES[which]
    ltypes = [int, float, np.int_, np.float_]
    c0 = (type(res) in ltypes
          or (hasattr(res, "__iter__")
              and len(res) == dres[which]
              and all([type(ds) in ltypes for ds in res])))
    if not c0:
        msg = ("Arg res must be either:\n"
               + "\t- float: unique resolution for all directions\n"
               + "\t- iterable of {} floats\n".format(dres[which])
               + "  You provided:\n{}".format(res))
        raise Exception(msg)
    if type(res) in ltypes:
        if which != 'edge':
            res = [float(res) for ii in range(dres[which])]
    else:
        if which == 'edge':
            msg = ("For edge, res cannot be an iterable!\n"
                   + "\t- res: {}".format(res))
            raise Exception(msg)
        res = [float(res[ii]) for ii in range(dres[which])]

    # domain (i.e.: sub-domain to be sampled, defined by its limits)
    ddomain = {'edge': 2, 'cross': 2, 'surface': 3, 'volume': 3}
    if domain is None:
        domain = [None for ii in range(ddomain[which])]
    c0 = (hasattr(domain, "__iter__")
          and len(domain) == ddomain[which]
          and all([dd is None
                   or (hasattr(dd, "__iter__")
                       and len(dd) == 2
                       and all([ss is None
                                or type(ss) in ltypes
                                for ss in dd]))
                   for dd in domain]))
    if not c0:
        msg = ("Arg domain must be a len()={} iterable".format(ddomain[which])
               + " where each element can be:\n"
               + "\t- an iterable 2 floats: [lower, upper] bounds\n"
               + "\t- None: no bounds\n"
               + "  You provided:\n{}".format(domain))
        raise Exception(msg)
    for ii in range(len(domain)):
        if domain[ii] is not None:
            domain[ii] = [float(domain[ii][0])
                          if domain[ii][0] is not None else None,
                          float(domain[ii][1])
                          if domain[ii][1] is not None else None]

    # resMode
    if resMode is None:
        resMode = _SAMPLE_RESMODE[which]
    c0 = isinstance(resMode, str) and resMode.lower() in ["abs", "rel"]
    if not c0:
        msg = ("Arg resMode must be in ['abs','rel']!\n"
               + "  You provided:\n{}".format(resMode))
        raise Exception(msg)
    resMode = resMode.lower()

    # ind (indices of points to be recovered)
    c0 = (ind is None
          or (isinstance(ind, np.ndarray)
              and ind.ndim == 1
              and ind.dtype == np.int_
              and np.all(ind >= 0))
          or (which == 'surface'
              and isinstance(ind, list)
              and all([isinstance(indi, np.ndarray)
                       and indi.ndim == 1
                       and indi.dtype == np.int_
                       and np.all(indi >= 0) for indi in ind])))
    if not c0:
        msg = ("Arg ind must be either:\n"
               + "\t- None: domain is used instead\n"
               + "\t- 1d np.ndarray of positive int: indices\n")
        if which == 'surface':
            msg += "\t- list of 1d np.ndarray of positive indices\n"
        msg += "  You provided:\n{}".format(ind)
        raise Exception(msg)

    return res, domain, resMode, ind


def _Ves_get_sampleEdge(
    VPoly,
    res=None,
    domain=None,
    resMode=None,
    offsetIn=0.0,
    VIn=None,
    margin=1.0e-9
):

    # -------------
    #  Check inputs

    # standard
    res, domain, resMode, ind = _Ves_get_sample_checkinputs(
        res=res,
        domain=domain,
        resMode=resMode,
        ind=None,
        which='edge',
    )

    # specific
    ltypes = [int, float, np.int_, np.float_]
    assert type(offsetIn) in ltypes

    # -------------
    #  Compute

    Pts, reseff, ind, N, Rref, VPolybis = _GG.discretize_vpoly(
        VPoly,
        float(res),
        mode=resMode,
        D1=domain[0],
        D2=domain[1],
        margin=margin,
        DIn=float(offsetIn),
        VIn=VIn,
    )
    return Pts, reseff, ind


def _Ves_get_sampleCross(
    VPoly,
    Min1,
    Max1,
    Min2,
    Max2,
    res=None,
    domain=None,
    resMode=None,
    ind=None,
    margin=1.0e-9,
    mode="flat",
):

    # -------------
    #  Check inputs

    # standard
    res, domain, resMode, ind = _Ves_get_sample_checkinputs(
        res=res,
        domain=domain,
        resMode=resMode,
        ind=ind,
        which='cross',
    )

    # specific
    assert mode in ["flat", "imshow"]

    # -------------
    #  Compute

    MinMax1 = np.array([Min1, Max1])
    MinMax2 = np.array([Min2, Max2])
    if ind is None:
        if mode == "flat":
            Pts, dS, ind, d1r, d2r = _GG.discretize_segment2d(
                MinMax1,
                MinMax2,
                res[0],
                res[1],
                D1=domain[0],
                D2=domain[1],
                mode=resMode,
                VPoly=VPoly,
                margin=margin,
            )
            out = (Pts, dS, ind, (d1r, d2r))
        else:
            x1, d1r, ind1, N1 = _GG._Ves_mesh_dlfromL_cython(
                MinMax1, res[0], domain[0], Lim=True,
                dLMode=resMode, margin=margin
            )
            x2, d2r, ind2, N2 = _GG._Ves_mesh_dlfromL_cython(
                MinMax2, res[1], domain[1], Lim=True,
                dLMode=resMode, margin=margin
            )
            xx1, xx2 = np.meshgrid(x1, x2)
            pts = np.squeeze([xx1, xx2])
            extent = (
                x1[0] - d1r / 2.0,
                x1[-1] + d1r / 2.0,
                x2[0] - d2r / 2.0,
                x2[-1] + d2r / 2.0,
            )
            out = (pts, x1, x2, extent)
    else:
        assert mode == "flat"
        c0 = type(ind) is np.ndarray and ind.ndim == 1
        c0 = c0 and ind.dtype in ["int32", "int64"] and np.all(ind >= 0)
        assert c0, "Arg ind must be a np.ndarray of int !"
        Pts, dS, d1r, d2r = _GG._Ves_meshCross_FromInd(
            MinMax1,
            MinMax2,
            res[0],
            res[1],
            ind,
            dSMode=resMode,
            margin=margin,
        )
        out = (Pts, dS, ind, (d1r, d2r))
    return out


def _Ves_get_sampleS(
    VPoly,
    res=None,
    domain=None,
    resMode="abs",
    ind=None,
    offsetIn=0.0,
    VIn=None,
    VType="Tor",
    VLim=None,
    nVLim=None,
    returnas="(X,Y,Z)",
    margin=1.0e-9,
    Multi=False,
    Ind=None,
):
    """ Sample the surface """

    # -------------
    #  Check inputs

    # standard
    res, domain, resMode, ind = _Ves_get_sample_checkinputs(
        res=res,
        domain=domain,
        resMode=resMode,
        ind=ind,
        which='surface',
    )

    # nVLim and VLim
    if not (type(nVLim) in [int, np.int_] and nVLim >= 0):
        msg = ("Arg nVLim must be a positive int\\n"
               + "  You provided:\n{} ({})".format(nVLim, type(nVLim)))
        raise Exception(msg)
    VLim = None if (VLim is None or nVLim == 0) else np.array(VLim)

    if not isinstance(Multi, bool):
        msg = ("Arg Multi must be a bool!\n"
               + "  You provided:\n{}".format(Multi))
        raise Exception(msg)

    # Check if Multi
    if nVLim > 1:
        assert VLim is not None, "For multiple Struct, Lim cannot be None !"
        assert all([hasattr(ll, "__iter__") and len(ll) == 2 for ll in VLim])
        if Ind is None:
            Ind = np.arange(0, nVLim)
        else:
            Ind = [Ind] if not hasattr(Ind, "__iter__") else Ind
            Ind = np.asarray(Ind).astype(int)
        if ind is not None:
            if isinstance(ind, np.ndarray):
                ind = [ind for ii in range(len(Ind))]
            elif not (isinstance(ind, list) and len(ind) == len(Ind)):
                msg = ("Arg ind must be a list of same len() as Ind!\n"
                       + "\t- provided: {}".format(ind))
                raise Exception(msg)

    else:
        VLim = [None] if VLim is None else [VLim.ravel()]
        if ind is not None:
            if not isinstance(ind, np.ndarray):
                msg = ("ind must be a np.ndarray if nVLim == 1\n"
                       + "\t- provided: {}".format(ind))
                raise Exception(msg)
        Ind = [0]

    if ind is None:
        pts, dS, ind, reseff = (
            [0 for ii in Ind],
            [0 for ii in Ind],
            [0 for ii in Ind],
            [[0, 0] for ii in Ind],
        )
        if VType.lower() == "tor":
            for ii in range(0, len(Ind)):
                if VLim[Ind[ii]] is None:
                    (pts[ii],
                     dS[ii],
                     ind[ii],
                     NL,
                     reseff[ii][0],
                     Rref,
                     reseff[ii][1],
                     nRPhi0,
                     VPbis) = _GG._Ves_Smesh_Tor_SubFromD_cython(
                         res[0],
                         res[1],
                         VPoly,
                         DR=domain[0],
                         DZ=domain[1],
                         DPhi=domain[2],
                         DIn=offsetIn,
                         VIn=VIn,
                         PhiMinMax=None,
                         Out=returnas,
                         margin=margin,
                     )
                else:
                    (pts[ii],
                     dS[ii],
                     ind[ii],
                     NL,
                     reseff[ii][0],
                     Rref,
                     dR0r,
                     dZ0r,
                     reseff[ii][1],
                     VPbis) = _GG._Ves_Smesh_TorStruct_SubFromD_cython(
                         VLim[Ind[ii]],
                         res[0],
                         res[1],
                         VPoly,
                         DR=domain[0],
                         DZ=domain[1],
                         DPhi=domain[2],
                         DIn=offsetIn,
                         VIn=VIn,
                         Out=returnas,
                         margin=margin,
                     )
                    reseff[ii] += [dR0r, dZ0r]
        else:
            for ii in range(0, len(Ind)):
                (pts[ii],
                 dS[ii],
                 ind[ii],
                 NL,
                 reseff[ii][0],
                 Rref,
                 reseff[ii][1],
                 dY0r,
                 dZ0r,
                 VPbis) = _GG._Ves_Smesh_Lin_SubFromD_cython(
                     VLim[Ind[ii]],
                     res[0],
                     res[1],
                     VPoly,
                     DX=domain[0],
                     DY=domain[1],
                     DZ=domain[2],
                     DIn=offsetIn,
                     VIn=VIn,
                     margin=margin,
                 )
                reseff[ii] += [dY0r, dZ0r]
    else:
        ind = ind if Multi else [ind]
        pts, dS, reseff = (
            [np.ones((3, 0)) for ii in Ind],
            [0 for ii in Ind],
            [[0, 0] for ii in Ind],
        )
        if VType.lower() == "tor":
            for ii in range(0, len(Ind)):
                if ind[Ind[ii]].size > 0:
                    if VLim[Ind[ii]] is None:
                        out_loc = _GG._Ves_Smesh_Tor_SubFromInd_cython(
                            res[0],
                            res[1],
                            VPoly,
                            ind[Ind[ii]],
                            DIn=offsetIn,
                            VIn=VIn,
                            PhiMinMax=None,
                            Out=returnas,
                            margin=margin,
                        )
                        pts[ii], dS[ii], NL, reseff[ii][0], Rref = out_loc[:5]
                        reseff[ii][1], nRPhi0, VPbis = out_loc[5:]
                    else:
                        out_loc = _GG._Ves_Smesh_TorStruct_SubFromInd_cython(
                            VLim[Ind[ii]],
                            res[0],
                            res[1],
                            VPoly,
                            ind[Ind[ii]],
                            DIn=offsetIn,
                            VIn=VIn,
                            Out=returnas,
                            margin=margin,
                        )
                        pts[ii], dS[ii], NL, reseff[ii][0], Rref = out_loc[:5]
                        dR0r, dZ0r, reseff[ii][1], VPbis = out_loc[5:]
                        reseff[ii] += [dR0r, dZ0r]
        else:
            for ii in range(0, len(Ind)):
                if ind[Ind[ii]].size > 0:
                    out_loc = _GG._Ves_Smesh_Lin_SubFromInd_cython(
                        VLim[Ind[ii]],
                        res[0],
                        res[1],
                        VPoly,
                        ind[Ind[ii]],
                        DIn=offsetIn,
                        VIn=VIn,
                        margin=margin,
                    )
                    pts[ii], dS[ii], NL, reseff[ii][0], Rref = out_loc[:5]
                    reseff[ii][1], dY0r, dZ0r, VPbis = out_loc[5:]
                    reseff[ii] += [dY0r, dZ0r]

    if len(VLim) == 1:
        pts, dS, ind, reseff = pts[0], dS[0], ind[0], reseff[0]
    return pts, dS, ind, reseff


def _Ves_get_sampleV(
    VPoly,
    Min1,
    Max1,
    Min2,
    Max2,
    res=None,
    domain=None,
    resMode=None,
    ind=None,
    VType="Tor",
    VLim=None,
    returnas="(X,Y,Z)",
    margin=1.0e-9,
    algo="new",
    num_threads=48,
):
    """ Sample the volume """

    # -------------
    #  Check inputs
    res, domain, resMode, ind = _Ves_get_sample_checkinputs(
        res=res,
        domain=domain,
        resMode=resMode,
        ind=ind,
        which='volume',
    )

    # ------------
    # Computation

    MinMax1 = np.array([Min1, Max1])
    MinMax2 = np.array([Min2, Max2])
    VLim = None if VType.lower() == "tor" else np.array(VLim).ravel()
    reseff = [None, None, None]
    if ind is None:
        if VType.lower() == "tor":
            if algo.lower() == "new":
                (pts, dV, ind,
                 reseff[0],
                 reseff[1],
                 reseff[2]) = _GG._Ves_Vmesh_Tor_SubFromD_cython(
                    res[0],
                    res[1],
                    res[2],
                    MinMax1,
                    MinMax2,
                    DR=domain[0],
                    DZ=domain[1],
                    DPhi=domain[2],
                    VPoly=VPoly,
                    Out=returnas,
                    margin=margin,
                    num_threads=num_threads,
                )
            else:
                (pts, dV, ind,
                 reseff[0],
                 reseff[1],
                 reseff[2]) = _GG._Ves_Vmesh_Tor_SubFromD_cython_old(
                    res[0],
                    res[1],
                    res[2],
                    MinMax1,
                    MinMax2,
                    DR=domain[0],
                    DZ=domain[1],
                    DPhi=domain[2],
                    VPoly=VPoly,
                    Out=returnas,
                    margin=margin,
                )

        else:
            (pts, dV, ind,
             reseff[0],
             reseff[1],
             reseff[2]) = _GG._Ves_Vmesh_Lin_SubFromD_cython(
                res[0],
                res[1],
                res[2],
                VLim,
                MinMax1,
                MinMax2,
                DX=domain[0],
                DY=domain[1],
                DZ=domain[2],
                VPoly=VPoly,
                margin=margin,
            )
    else:
        if VType.lower() == "tor":
            if algo.lower() == "new":
                (pts, dV,
                 reseff[0],
                 reseff[1],
                 reseff[2]) = _GG._Ves_Vmesh_Tor_SubFromInd_cython(
                    res[0],
                    res[1],
                    res[2],
                    MinMax1,
                    MinMax2,
                    ind,
                    Out=returnas,
                    margin=margin,
                    num_threads=num_threads,
                )
            else:
                (pts, dV,
                 reseff[0],
                 reseff[1],
                 reseff[2]) = _GG._Ves_Vmesh_Tor_SubFromInd_cython_old(
                    res[0],
                    res[1],
                    res[2],
                    MinMax1,
                    MinMax2,
                    ind,
                    Out=returnas,
                    margin=margin,
                )
        else:
            (pts, dV,
             reseff[0],
             reseff[1],
             reseff[2]) = _GG._Ves_Vmesh_Lin_SubFromInd_cython(
                 res[0],
                 res[1],
                 res[2],
                 VLim,
                 MinMax1,
                 MinMax2,
                 ind,
                 margin=margin
             )
    return pts, dV, ind, reseff


# ==============================================================================
# =  phi / theta projections for magfieldlines
# ==============================================================================


def _Struct_get_phithetaproj(ax=None, poly_closed=None, lim=None, noccur=0):

    # phi = toroidal angle
    if noccur == 0:
        Dphi = np.array([[-np.pi, np.pi]])
        nphi = np.r_[1]
    else:
        assert lim.ndim == 2, str(lim)
        nphi = np.ones((noccur,), dtype=int)
        ind = (lim[:, 0] > lim[:, 1]).nonzero()[0]
        Dphi = np.concatenate((lim, np.full((noccur, 2), np.nan)), axis=1)
        if ind.size > 0:
            for ii in ind:
                Dphi[ii, :] = [lim[ii, 0], np.pi, -np.pi, lim[ii, 1]]
                nphi[ii] = 2

    # theta = poloidal angle
    Dtheta = np.arctan2(poly_closed[1, :] - ax[1], poly_closed[0, :] - ax[0])
    Dtheta = np.r_[np.min(Dtheta), np.max(Dtheta)]
    if Dtheta[0] > Dtheta[1]:
        ntheta = 2
        Dtheta = [Dtheta[0], np.pi, -np.pi, Dtheta[1]]
    else:
        ntheta = 1

    return nphi, Dphi, ntheta, Dtheta


def _get_phithetaproj_dist(
    poly_closed,
    ax,
    Dtheta,
    nDtheta,
    Dphi,
    nDphi,
    theta,
    phi,
    ntheta,
    nphi,
    noccur,
):

    if nDtheta == 1:
        ind = (theta >= Dtheta[0]) & (theta <= Dtheta[1])
    else:
        ind = (theta >= Dtheta[0]) | (theta <= Dtheta[1])

    disttheta = np.full((theta.size,), np.nan)

    # phi within Dphi
    if noccur > 0:
        indphi = np.zeros((nphi,), dtype=bool)
        for ii in range(0, noccur):
            for jj in range(0, nDphi[ii]):
                indphi |= (phi >= Dphi[ii, jj]) & (phi <= Dphi[ii, jj + 1])
        if not np.any(indphi):
            return disttheta, indphi
    else:
        indphi = np.ones((nphi,), dtype=bool)

    # No theta within Dtheta
    if not np.any(ind):
        return disttheta, indphi

    # Check for non-parallel AB / u pairs
    u = np.array([np.cos(theta), np.sin(theta)])
    AB = np.diff(poly_closed, axis=1)
    detABu = AB[0, :, None] * u[1, None, :] - AB[1, :, None] * u[0, None, :]
    inddet = ind[None, :] & (np.abs(detABu) > 1.0e-9)
    if not np.any(inddet):
        return disttheta, indphi

    nseg = poly_closed.shape[1] - 1
    k = np.full((nseg, ntheta), np.nan)

    OA = poly_closed[:, :-1] - ax[:, None]
    detOAu = (OA[0, :, None] * u[1, None, :] - OA[1, :, None] * u[0, None, :])[
        inddet
    ]
    ss = -detOAu / detABu[inddet]
    inds = (ss >= 0.0) & (ss < 1.0)
    inddet[inddet] = inds

    if not np.any(inds):
        return disttheta, indphi

    scaOAu = (OA[0, :, None] * u[0, None, :] + OA[1, :, None] * u[1, None, :])[
        inddet
    ]
    scaABu = (AB[0, :, None] * u[0, None, :] + AB[1, :, None] * u[1, None, :])[
        inddet
    ]
    k[inddet] = scaOAu + ss[inds] * scaABu
    indk = k[inddet] > 0.0
    inddet[inddet] = indk

    if not np.any(indk):
        return disttheta, indphi

    k[~inddet] = np.nan
    indok = np.any(inddet, axis=0)
    disttheta[indok] = np.nanmin(k[:, indok], axis=0)

    return disttheta, indphi


# ==============================================================================
# =  LOS functions
# ==============================================================================
def LOS_PRMin(Ds, us, kOut=None, Eps=1.0e-12, squeeze=True, Test=True):
    """  Compute the point on the LOS where the major radius is minimum """
    if Test:
        assert Ds.ndim in [1, 2, 3] and 3 in Ds.shape and Ds.shape == us.shape
    if kOut is not None:
        kOut = np.atleast_1d(kOut)
        assert kOut.size == Ds.size / 3

    if Ds.ndim == 1:
        Ds, us = Ds[:, None, None], us[:, None, None]
    elif Ds.ndim == 2:
        Ds, us = Ds[:, :, None], us[:, :, None]
    if kOut is not None:
        if kOut.ndim == 1:
            kOut = kOut[:, None]
    _, nlos, nref = Ds.shape

    kRMin = np.full((nlos, nref), np.nan)
    uparN = np.sqrt(us[0, :, :] ** 2 + us[1, :, :] ** 2)

    # Case with u vertical
    ind = uparN > Eps
    kRMin[~ind] = 0.0

    # Else
    kRMin[ind] = (
        -(us[0, ind] * Ds[0, ind] + us[1, ind] * Ds[1, ind]) / uparN[ind] ** 2
    )

    # Check
    kRMin[kRMin <= 0.0] = 0.0
    if kOut is not None:
        kRMin[kRMin > kOut] = kOut[kRMin > kOut]

    # squeeze
    if squeeze:
        if nref == 1 and nlos == 11:
            kRMin = kRMin[0, 0]
        elif nref == 1:
            kRMin = kRMin[:, 0]
        elif nlos == 1:
            kRMin = kRMin[0, :]
    return kRMin


def LOS_CrossProj(
    VType,
    Ds,
    us,
    kOuts,
    proj="All",
    multi=False,
    num_threads=16,
    return_pts=False,
    Test=True,
):
    """ Compute the parameters to plot the poloidal projection of the LOS  """
    assert type(VType) is str and VType.lower() in ["tor", "lin"]
    dproj = {
        "cross": ("R", "Z"),
        "hor": ("x,y"),
        "all": ("R", "Z", "x", "y"),
        "3d": ("x", "y", "z"),
    }
    assert type(proj) in [str, tuple]
    if type(proj) is tuple:
        assert all([type(pp) is str for pp in proj])
        lcoords = proj
    else:
        proj = proj.lower()
        assert proj in dproj.keys()
        lcoords = dproj[proj]
    if return_pts:
        assert proj in ["cross", "hor", "3d"]

    lc = [Ds.ndim == 3, Ds.shape == us.shape]
    if not all(lc):
        msg = "Ds and us must have the same shape and dim in [2,3]:\n"
        msg += "    - provided Ds.shape: %s\n" % str(Ds.shape)
        msg += "    - provided us.shape: %s" % str(us.shape)
        raise Exception(msg)
    lc = [kOuts.size == Ds.size / 3, kOuts.shape == Ds.shape[1:]]
    if not all(lc):
        msg = "kOuts must have the same shape and ndim = Ds.ndim-1:\n"
        msg += "    - Ds.shape    : %s\n" % str(Ds.shape)
        msg += "    - kOutss.shape: %s" % str(kOuts.shape)
        raise Exception(msg)

    # Prepare inputs
    _, nlos, nseg = Ds.shape

    # Detailed sampling for 'tor' and ('cross' or 'all')
    R, Z = None, None
    if "R" in lcoords or "Z" in lcoords:
        angcross = np.arccos(
            np.sqrt(us[0, ...] ** 2 + us[1, ...] ** 2)
            / np.sqrt(np.sum(us ** 2, axis=0))
        )
        resnk = np.ceil(25.0 * (1 - (angcross / (np.pi / 4) - 1) ** 2) + 5)
        resnk = 1.0 / resnk.ravel()

        # Use optimized get sample
        DL = np.vstack((np.zeros((nlos * nseg,), dtype=float), kOuts.ravel()))

        k, reseff, lind = _GG.LOS_get_sample(
            nlos * nseg,
            resnk,
            DL,
            dmethod="rel",
            method="simps",
            num_threads=num_threads,
            Test=Test,
        )

        assert lind.size == nseg * nlos - 1
        ind = lind[nseg - 1 :: nseg]  # noqa
        nbrep = np.r_[lind[0], np.diff(lind), k.size - lind[-1]]
        pts = np.repeat(Ds.reshape((3, nlos * nseg)), nbrep, axis=1) + k[
            None, :
        ] * np.repeat(us.reshape((3, nlos * nseg)), nbrep, axis=1)

        if return_pts:
            pts = np.array([np.hypot(pts[0, :], pts[1, :]), pts[2, :]])
            if multi:
                pts = np.split(pts, ind, axis=1)
            else:
                pts = np.insert(pts, ind, np.nan, axis=1)
        else:
            if multi:
                if "R" in lcoords:
                    R = np.split(np.hypot(pts[0, :], pts[1, :]), ind)
                if "Z" in lcoords:
                    Z = np.split(pts[2, :], ind)
            else:
                if "R" in lcoords:
                    R = np.insert(np.hypot(pts[0, :], pts[1, :]), ind, np.nan)
                if "Z" in lcoords:
                    Z = np.insert(pts[2, :], ind, np.nan)

    # Normal sampling => pts
    # unnecessary only if 'tor' and 'cross'
    x, y, z = None, None, None
    if "x" in lcoords or "y" in lcoords or "z" in lcoords:
        pts = np.concatenate(
            (Ds, Ds[:, :, -1:] + kOuts[None, :, -1:] * us[:, :, -1:]), axis=-1
        )

        if multi:
            ind = np.arange(1, nlos) * (nseg + 1)
            pts = pts.reshape((3, nlos * (nseg + 1)))
        else:
            nancoords = np.full((3, nlos, 1), np.nan)
            pts = np.concatenate((pts, nancoords), axis=-1)
            pts = pts.reshape((3, nlos * (nseg + 2)))

        if return_pts:
            assert proj in ["hor", "3d"]
            if multi:
                if proj == "hor":
                    pts = np.split(pts[:2, :], ind, axis=1)
                else:
                    pts = np.split(pts, ind, axis=1)
            elif proj == "hor":
                pts = pts[:2, :]

        else:
            if multi:
                if "x" in lcoords:
                    x = np.split(pts[0, :], ind)
                if "y" in lcoords:
                    y = np.split(pts[1, :], ind)
                if "z" in lcoords:
                    z = np.split(pts[2, :], ind)
            else:
                if "x" in lcoords:
                    x = pts[0, :]
                if "y" in lcoords:
                    y = pts[1, :]
                if "z" in lcoords:
                    z = pts[2, :]

    if return_pts:
        return pts
    else:
        return R, Z, x, y, z


# ==============================================================================
# =  Meshing & signal
# ==============================================================================
def LOS_get_sample(D, u, dL, DL=None, dLMode="abs", method="sum", Test=True):
    """ Return the sampled line, with the specified method

    'linspace': return the N+1 edges, including the first and last point
    'sum' : return the N middle of the segments
    'simps': return the N+1 egdes, where N has to be even
             (scipy.simpson requires an even number of intervals)
    'romb' : return the N+1 edges, where N+1 = 2**k+1
             (fed to scipy.romb for integration)
    """
    if Test:
        assert all(
            [type(dd) is np.ndarray and dd.shape == (3,) for dd in [D, u]]
        )
        assert not hasattr(dL, "__iter__")
        assert DL is None or all(
            [
                hasattr(DL, "__iter__"),
                len(DL) == 2,
                all([not hasattr(dd, "__iter__") for dd in DL]),
            ]
        )
        assert dLMode in ["abs", "rel"]
        assert type(method) is str and method in [
            "linspace",
            "sum",
            "simps",
            "romb",
        ]
    # Compute the min number of intervals to satisfy the specified resolution
    N = (
        int(np.ceil((DL[1] - DL[0]) / dL))
        if dLMode == "abs"
        else int(np.ceil(1.0 / dL))
    )
    # Modify N according to the desired method
    if method == "simps":
        N = N if N % 2 == 0 else N + 1
    elif method == "romb":
        N = 2 ** int(np.ceil(np.log(N) / np.log(2.0)))

    # Derive k and dLr
    if method == "sum":
        dLr = (DL[1] - DL[0]) / N
        k = DL[0] + (0.5 + np.arange(0, N)) * dLr
    else:
        k, dLr = np.linspace(
            DL[0], DL[1], N + 1, endpoint=True, retstep=True, dtype=float
        )

    Pts = D[:, np.newaxis] + k[np.newaxis, :] * u[:, np.newaxis]
    return Pts, k, dLr


def LOS_calc_signal(
    ff, D, u, dL, DL=None, dLMode="abs", method="romb", Test=True
):
    assert hasattr(ff, "__call__"), (
        "Arg ff must be a callable (function) taking at least 1 positional ",
        "Pts (a (3,N) np.ndarray of cartesian (X,Y,Z) coordinates) !",
    )
    assert not method == "linspace"
    Pts, k, dLr = LOS_get_sample(
        D, u, dL, DL=DL, dLMode=dLMode, method=method, Test=Test
    )
    out = insp(ff)
    N = np.sum(
        [
            (
                pp.kind == pp.POSITIONAL_OR_KEYWORD
                and pp.default is pp.empty
            )
            for pp in out.parameters.values()
        ]
    )

    if N == 1:
        Vals = ff(Pts)
    elif N == 2:
        Vals = ff(Pts, np.tile(-u, (Pts.shape[1], 1)).T)
    else:
        raise ValueError(
            "The function (ff) assessing the emissivity locally "
            + "must take a single positional argument: Pts a (3,N)"
            + " np.ndarray of (X,Y,Z) cartesian coordinates !"
        )

    Vals[np.isnan(Vals)] = 0.0
    if method == "sum":
        Int = np.sum(Vals) * dLr
    elif method == "simps":
        Int = scpintg.simps(Vals, x=None, dx=dLr)
    elif method == "romb":
        Int = scpintg.romb(Vals, dx=dLr, show=False)
    return Int


# ==============================================================================
# =  Solid Angle particle
# ==============================================================================


def calc_solidangle_particle(
    traj, pts, r=1.0, config=None, approx=True, aniso=False, block=True
):
    """ Compute the solid angle subtended by a particle along a trajectory

    The particle has radius r, and trajectory (array of points) traj
    It is observed from pts (array of points)

    traj and pts are (3,N) and (3,M) arrays of cartesian coordinates

    approx = True => use approximation
    aniso = True => return also unit vector of emission
    block = True consider LOS collisions (with Ves, Struct...)

    if block:
        config = config used for LOS collisions

    Return:
    -------
    sang: np.ndarray
        (N,M) Array of floats, solid angles

    """
    ################
    # Prepare inputs
    traj = np.ascontiguousarray(traj, dtype=float)
    pts = np.ascontiguousarray(pts, dtype=float)
    r = np.r_[r].astype(float).ravel()

    # Check booleans
    assert type(approx) is bool
    assert type(aniso) is bool
    assert type(block) is bool

    # Check config
    assert config is None or config.__class__.__name__ == "Config"
    assert block == (config is not None)

    # Check pts, traj and r are array of good shape
    assert traj.ndim in [1, 2]
    assert pts.ndim in [1, 2]
    assert 3 in traj.shape and 3 in pts.shape
    if traj.ndim == 1:
        traj = traj.reshape((3, 1))
    if traj.shape[0] != 3:
        traj = traj.T
    if pts.ndim == 1:
        pts = pts.reshape((3, 1))
    if pts.shape[0] != 3:
        pts = pts.T

    # get npart
    ntraj = traj.shape[1]
    nr = r.size

    npart = max(nr, ntraj)
    assert nr in [1, npart]
    assert ntraj in [1, npart]
    if nr < npart:
        r = np.full((npart,), r[0])
    if ntraj < npart:
        traj = np.repeat(traj, npart, axis=1)

    ################
    # Main computation

    # traj2pts vector, with length (3d array (3,N,M))
    vect = pts[:, None, :] - traj[:, :, None]
    len_v = np.sqrt(np.sum(vect ** 2, axis=0))

    # If aniso or block, normalize
    if aniso or block:
        vect = vect / len_v[None, :, :]

    # Solid angle
    if approx:
        sang = np.pi * r[None, :] ** 2 / len_v ** 2
    else:
        sang = 2.0 * np.pi * (1 - np.sqrt(1.0 - r ** 2[None, :] / len_v ** 2))

    # block
    if block:
        kwdargs = config._get_kwdargs_LOS_isVis()
        # TODO : modify this function along issue #102
        indnan = _GG.LOS_areVis_PtsFromPts_VesStruct(
            traj, pts, k=len_v, vis=False, **kwdargs
        )
        sang[indnan] = 0.0
        vect[indnan, :] = np.nan

    ################
    # Return

    if aniso:
        return sang, vect
    else:
        return sang


def calc_solidangle_particle_integ(
    traj, r=1.0, config=None, approx=True, block=True, res=0.01
):

    # step0: if block : generate kwdargs from config

    # step 1: sample cross-section

    # step 2: loop on R of  pts of cross-section (parallelize ?)
    # => fix nb. of phi for the rest of the loop

    # loop of Z

    # step 3: loop phi
    # Check visibility (if block = True) for each phi (LOS collision)
    # If visible => compute solid angle
    # integrate (sum * res) on each phi the solid angle

    # Return sang as (N,nR,nZ) array
    return



#!/usr/bin/env python

# Built-in
import sys
import os
from shutil import copyfile


_HERE = os.path.abspath(os.path.dirname(__file__))


# import parser dict
sys.path.insert(1, _HERE)
from _dparser import _DPARSER
_ = sys.path.pop(1)


###################################################
###################################################
#       function
###################################################


def custom(target=None, source=None,
           files=None, directories=None, ddef=None):

    # --------------
    # Check inputs
    kwd = locals()
    for k0 in set(ddef.keys()).intersection(kwd.keys()):
        if kwd[k0] is None:
            kwd[k0] = ddef[k0]
    target, source = kwd['target'], kwd['source']
    files, directories = kwd['files'], kwd['directories']

    # Caveat (up to now only relevant for _TARGET)
    if target != ddef['target']:
        msg = ""
        raise Exception(msg)

    # Check files
    if isinstance(files, str):
        files = [files]
    c0 = (not isinstance(files, list)
          or any([ff not in ddef['files'] for ff in files]))
    if c0 is True:
        msg = "All files should be in {}".format(ddef['files'])
        raise Exception(msg)

    # Try creating directory and copying modules
    try:
        # Create .tofu/ if non-existent
        if not os.path.isdir(target):
            os.mkdir(target)

        # Create directories
        for dd in directories:
            if not os.path.isdir(os.path.join(target, dd)):
                os.mkdir(os.path.join(target, dd))

        # Copy files
        for ff in files:
            mod, f0 = ff.split('_')[1:]
            copyfile(os.path.join(source, mod, '_'+f0),
                     os.path.join(target, ff))

        msg = ("A local copy of default tofu parameters is now in:\n"
               + "\t{}/\n".format(target)
               + "You can edit it to spice up your tofu")
        print(msg)

    except Exception as err:
        msg = (str(err) + '\n\n'
               + "A problem occured\n"
               + "tofu-custom tried to create a local directory .tofu/ in "
               + "your home {}\n".format(target)
               + "But it could not, check the error message above to debug\n"
               + "Most frequent cause is a permission issue")
        raise Exception(msg)


###################################################
###################################################
#       bash call (main)
###################################################


def main():
    # Parse input arguments
    # Instanciate parser
    ddef, parser = _DPARSER['custom']()

    # Parse arguments
    args = parser.parse_args()

    # Call function
    custom(ddef=ddef, **dict(args._get_kwargs()))


if __name__ == '__main__':
    main()




import sys
import os
import scipy.io as scpio


import numpy as np


_HERE = os.path.abspath(os.path.dirname(__file__))
_TOFU_PATH = os.path.abspath(os.path.join(_HERE, '..', '..', '..'))


# import parser dict
sys.path.insert(1, _TOFU_PATH)
import tofu.geom as tfg
_ = sys.path.pop(1)


def get_from_mat_file_from_tcv_modified(pfe=None, Exp='TCV'):

    import tofu.geom as tfg

    # Check inputs
    c0 = (os.path.isfile(pfe)
          and pfe[-4:] == '.mat')
    if c0 is not True:
        msg = ("File does not exist or not a .mat file!\n"
               + "\t- pfe: {}\n".format(pfe))
        raise Exception(msg)

    # Extract data
    out = scpio.loadmat(pfe)

    # Check file conformity
    lka = ['h2']
    lkb = [k0 for k0 in out.keys() if '__' not in k0]
    if lka != lkb:
        msg = ("Content of file not as expected!\n"
               + "\t- pfe: {}\n".format(pfe)
               + "\t- expected keys: {}\n".format(lka)
               + "\t- observed keys: {}".format(lkb))
        raise Exception(msg)

    # Create objects
    ls = []
    out = out[lka[0]]
    for ii in range(out.shape[1]):
        name = out[0][ii][0][0].replace('_', '').replace('-', '')
        if name == 'Bcoil':
            continue
        if name[0] in ['A', 'B', 'C', 'D', 'E', 'F', 'T']:
            poly = out[0][ii][1].T
            obj = tfg.CoilPF(Name=name, Poly=poly, Exp=Exp)
        elif name == 'v':
            if 512 not in out[0][ii][1].shape:
                msg = ("Size of {} not as expected!\n".format(name)
                       + "\t- expected: 512 pts\n"
                       + "\t- observed: {} pts".format(out[0][ii][1].size/2))
                raise Exception(msg)
            iin = np.r_[0, np.arange(512)]
            iout = np.arange(1, 257)
            poly = out[0][ii][1][iin, :].T
            obj = tfg.Ves(Name=name, Poly=poly, Exp=Exp)
        elif name == 't':
            if 314 not in out[0][ii][1].shape:
                msg = ("Size of {} not as expected!\n".format(name)
                       + "\t- expected: 314 pts\n"
                       + "\t- observed: {} pts".format(out[0][ii][1].size/2))
                raise Exception(msg)
            ind = np.arange(57)
            poly = out[0][ii][1][ind, :].T
            obj = tfg.Ves(Name=name, Poly=poly, Exp=Exp)
        else:
            msg = "Unidentified element {}".format(name)
            raise Exception(msg)
        ls.append(obj)
    return tfg.Config(lStruct=ls, Name='V1', Exp=Exp)



"""
This module stores all the default setting of ToFu
Including in particular computing parameters, dictionnaries and figures
"""

#import matplotlib
#matplotlib.use('WxAgg')
#matplotlib.interactive(True)

import matplotlib.pyplot as plt
from matplotlib.path import Path
from mpl_toolkits.mplot3d import Axes3D
import numpy as np

#from mayavi import mlab
import datetime as dtm
import time as time


"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings generic
###############################################################################
###############################################################################
"""

# Default saving Paths
KeyRP = '/ToFu/src'
SaveAddIn = '/Objects_AUG/'
SaveAddObj = '/Objects_AUG/'
SaveAddOut = '/Outputs_AUG/'
MeshSaveAdd = '/Objects/'
dtmFormat = "D%Y%m%d_T%H%M%S"



AllowedExp = [None,'AUG','MISTRAL','JET','ITER','TCV','TS','WEST','KSTAR','Misc','Test']

"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings of ToFu_Geom
###############################################################################
###############################################################################
"""


# ---------- Common to several classes ----------

Legpropd = {'size':10}
TorLegd = {'frameon':False,'ncol':1,'bbox_to_anchor':(1.01, 1),'loc':2,'borderaxespad':0.,'prop':Legpropd}


#####################################################################
########################  Ves class  ################################
#####################################################################


# ------------ Computing settings ---------------

TorNP = 50
TorRelOff = 0.05
TorInsideNP = 100
TorSplprms = [100.,2.,3]
DetBaryCylNP1 = 50
DetBaryCylNP2 = 200

# --- Plotting dictionaries and parameters ------

TorPd = {'c':'k','lw':2}
TorId = {'c':'k','ls':'dashed','marker':'x','markersize':8,'mew':2}
TorBsd = {'c':'b','ls':'dashed','marker':'x','markersize':8,'mew':2}
TorBvd = {'c':'g','ls':'dashed','marker':'x','markersize':8,'mew':2}
TorVind = {'color':'r','scale':10}
TorITord = {'c':'k','ls':'dashed'}
TorBsTord = {'c':'b','ls':'dashed'}
TorBvTord = {'c':'g','ls':'dashed'}
TorNTheta = 50
Tor3DThetalim = [np.pi/2,2*np.pi]
Tor3DThetamin = np.pi/20.
TorP3Dd = {'color':(0.8,0.8,0.8,1.),'rstride':1,'cstride':1,'linewidth':0, 'antialiased':False}
TorPFilld = {'edgecolor':(0.8,0.8,0.8,1.),'facecolor':(0.8,0.8,0.8,1.),'linestyle':'solid','linewidth':1}
TorPAng = 'theta'
TorPAngUnit = 'rad'
TorPSketch = True
TorP3DFilld = {'color':(0.8,0.8,0.8,0.4),'linestyle':'solid','linewidth':0}

Vesdict = dict(Lax=None, Proj='All', Elt='PIBsBvV', Pdict=None, Idict=TorId, Bsdict=TorBsd, Bvdict=TorBvd, Vdict=TorVind,
            IdictHor=TorITord, BsdictHor=TorBsTord, BvdictHor=TorBvTord, Lim=Tor3DThetalim, Nstep=TorNTheta, LegDict=TorLegd, draw=True, Test=True)


# -------------- Figures ------------------------


def Plot_LOSProj_DefAxes(Mode, Type='Tor', a4=False):
    assert Mode in ['Cross','Hor','All'], "Arg should be 'Cross' or 'Hor' or 'All' !"
    assert Type in ['Tor','Lin'], "Arg Type must be in ['Tor','Lin'] !"
    if Mode == 'Cross':
        fW,fH,fdpi,axCol = (6,8,80,'w') if not a4 else (8.27,11.69,80,'w')
        axPos = [0.15, 0.15, 0.6, 0.7]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
        if Type=='Tor':
            ax.set_xlabel(r"R (m)"),    ax.set_ylabel(r"Z (m)")
        else:
            ax.set_xlabel(r"Y (m)"), ax.set_ylabel(r"Z (m)")
        ax.set_aspect(aspect="equal", adjustable='datalim')
        return ax
    elif Mode == 'Hor':
        fW,fH,fdpi,axCol = (6,8,80,'w') if not a4 else (8.27,11.69,80,'w')
        axPos = [0.15, 0.15, 0.6, 0.7]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
        ax.set_xlabel(r"X (m)"),    ax.set_ylabel(r"Y (m)")
        ax.set_aspect(aspect="equal", adjustable='datalim')
        return ax
    elif Mode=='All':
        fW,fH,fdpi,axCol = (16,8,80,'w')  if not a4 else (11.69,8.27,80,'w')
        axPosP, axPosT = [0.07, 0.1, 0.3, 0.8], [0.55, 0.1, 0.3, 0.8]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        axP = f.add_axes(axPosP,frameon=True,axisbg=axCol)
        axT = f.add_axes(axPosT,frameon=True,axisbg=axCol)
        if Type=='Tor':
            axP.set_xlabel(r"R (m)"),   axP.set_ylabel(r"Z (m)")
        else:
            axP.set_xlabel(r"Y (m)"),   axP.set_ylabel(r"Z (m)")
        axT.set_xlabel(r"X (m)"),   axT.set_ylabel(r"Y (m)")
        axP.set_aspect(aspect="equal", adjustable='datalim')
        axT.set_aspect(aspect="equal", adjustable='datalim')
        return axP, axT

def Plot_3D_plt_Tor_DefAxes(a4=False):
    fW,fH,fdpi,axCol = (14,10,80,'w') if not a4 else (11.69,8.27,80,'w')
    axPos = [0.05, 0.05, 0.75, 0.85]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,axisbg=axCol,projection='3d')
    ax.set_xlabel(r"X (m)")
    ax.set_ylabel(r"Y (m)")
    ax.set_zlabel(r"Z (m)")
    ax.set_aspect(aspect="equal", adjustable='datalim')
    return ax


def Plot_Impact_DefAxes(Proj, Ang='theta', AngUnit='rad', a4=False, Sketch=True):
    if Proj == 'Cross':
        fW,fH,fdpi,axCol = (10,6,80,'w') if not a4 else (11.69,8.27,80,'w')
        axPos = [0.12, 0.12, 0.60, 0.8]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        ax, axSketch = f.add_axes(axPos,frameon=True,axisbg=axCol), []
        XAng = r"$\theta$" if Ang=='theta' else r"$\xi$"
        XUnit = r"$(rad.)$" if AngUnit=='rad' else r"$(deg.)$"
        XTickLab = [r"$0$",r"$\pi/4$",r"$\pi/2$",r"$3\pi/4$",r"$\pi$"] if AngUnit=='rad' else [r"$0$",r"$90$",r"$180$",r"$270$",r"$360$"]
        ax.set_xlabel(XAng+r" "+XUnit)
        ax.set_ylabel(r"$p$ $(m)$")
        ax.set_xlim(0,np.pi)
        ax.set_ylim(-1.5,1.5)
        ax.set_xticks(np.pi*np.array([0.,1./4.,1./2.,3./4.,1.]))
        ax.set_xticklabels(XTickLab)
        if Sketch:
            axSketch = f.add_axes([0.75, 0.10, 0.15, 0.15],frameon=False,axisbg=axCol)
            Pt, Line, Hor, theta, ksi = np.array([[0,-0.8],[0,0.8]]), np.array([[-1.6,0.1],[0,1.7]]), np.array([[-0.4,0.2],[1.2,1.2]]), np.linspace(0,3.*np.pi/4.,30), np.linspace(0,np.pi/4.,10)
            theta, ksi = np.array([0.3*np.cos(theta),0.3*np.sin(theta)]), np.array([-0.4+0.4*np.cos(ksi), 1.2+0.4*np.sin(ksi)])
            axSketch.plot(Pt[0,:],Pt[1,:],'+k',Pt[0,:],Pt[1,:],'--k',Line[0,:],Line[1,:],'-k', Hor[0,:],Hor[1,:],'-k', theta[0,:],theta[1,:],'-k', ksi[0,:],ksi[1,:],'-k')
            axSketch.annotate(r"$\theta$", xy=(0.3,0.4),xycoords='data',va="center", ha="center")
            axSketch.annotate(r"$\xi$", xy=(0.1,1.4),xycoords='data',va="center", ha="center")
            axSketch.annotate(r"$p$", xy=(-0.7,0.3),xycoords='data',va="center", ha="center")
            axSketch.set_xticks([]), axSketch.set_yticks([])
            axSketch.axis("equal")
        return ax, axSketch
    elif Proj.lower() == '3d':
        fW,fH,fdpi,axCol = (11,9,80,'w') if not a4 else (11.69,8.27,80,'w')
        axPos = [0.1, 0.1, 0.65, 0.8]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        ax = f.add_axes(axPos,axisbg=axCol,projection='3d')
        XAng = r"$\theta$" if Ang=='theta' else r"$\xi$"
        XUnit = r"$(rad.)$" if AngUnit=='rad' else r"$(deg.)$"
        XTickLab = [r"$0$",r"$\pi/4$",r"$\pi/2$",r"$3\pi/4$",r"$\pi$"] if AngUnit=='rad' else [r"$0$",r"$90$",r"$180$",r"$270$",r"$360$"]
        ax.set_xlabel(XAng+r" "+XUnit)
        ax.set_ylabel(r"$p$ $(m)$")
        ax.set_zlabel(r"$\phi$ $(rad)$")
        ax.set_xlim(0,np.pi)
        ax.set_ylim(-1.5,1.5)
        ax.set_zlim(-np.pi/2.,np.pi/2.)
        ax.set_xticks(np.pi*np.array([0.,1./4.,1./2.,3./4.,1.]))
        ax.set_xticklabels(XTickLab)
        return [ax]

#def Plot_3D_mlab_Tor_DefFig():
#    fW,fH,fBgC = 700,500,(1.,1.,1.)
#    axPosP, axPosT = [0.07, 0.1, 0.3, 0.8], [0.55, 0.1, 0.3, 0.8]
#    f = mlab.figure(bgcolor=fBgC,fgcolor=None,size=(fW,fH))
#    return f



#####################################################################
########################  Struct class  #############################
#####################################################################


# --- Plotting dictionaries and parameters ------

StructPd = {'edgecolor':'k','linewidth':1}
StructP3Dd = {'color':(0.8,0.8,0.8,1.),'rstride':1,'cstride':1,'linewidth':0, 'antialiased':False}

Vesdict = dict(Lax=None, Proj='All', Elt='PIBsBvV', Pdict=None, Idict=TorId, Bsdict=TorBsd, Bvdict=TorBvd, Vdict=TorVind,
            IdictHor=TorITord, BsdictHor=TorBsTord, BvdictHor=TorBvTord, Lim=Tor3DThetalim, Nstep=TorNTheta, LegDict=TorLegd, draw=True, Test=True)




#####################################################################
########################  LOS class  ################################
#####################################################################


# Number of points for plotting poloidal projection as a function of PolProjAng

def kpVsPolProjAng(x):
    return np.ceil(25.*(1 - (x/(np.pi/4)-1)**2) + 2)


# ------------ Computing settings ---------------
LOSDiscrtSLim = (0.,1,)
LOSDiscrtSLMode = 'norm'
LOSDiscrtDS = 0.005
LOSDiscrtSMode = 'm'


# --- Plotting dictionaries and parameters ------

LOSLd = {'c':'k','lw':2}
LOSMd = {'c':'k','ls':'None','lw':2,'marker':'x','markersize':8,'mew':2}
LOSMImpd = {'c':'k','ls':'None','lw':2,'marker':'x','markersize':8,'mew':2}
LOSLplot = 'Tot'
LOSImpAng = 'theta'
LOSImpAngUnit = 'rad'
LOSImpElt = 'LV'

LOSdict = dict(Lax=None, Proj='All', Lplot=LOSLplot, Elt='LDIORP', EltVes='', Leg='', Ldict=LOSLd, MdictD=LOSMd, MdictI=LOSMd, MdictO=LOSMd, MdictR=LOSMd, MdictP=LOSMd, LegDict=TorLegd, Vesdict=Vesdict, draw=True, Test=True)

# -------------- Figures ------------------------



#####################################################################
###################### Lens class  ################################
#####################################################################






# -------------- Figures ------------------------

def Plot_Lens_Alone_DefAxes(a4=False):
    axCol = 'w'
    (fW,fH) = (11.69,8.27) if a4 else (20,8)
    axPos = [0.05, 0.1, 0.9, 0.85]
    f = plt.figure(facecolor="w",figsize=(fW,fH))
    ax = f.add_axes(axPos,axisbg=axCol)
    ax.set_xlabel(r"x (m)")
    ax.set_ylabel(r"y (m)")
    return ax




#####################################################################
###################### Detect class  ################################
#####################################################################


# ------------ Computing settings ---------------

DetSpanRMinMargin = 0.9
DetSpanNEdge = 5
DetSpanNRad = 5

DetConeNEdge = 8
DetConeNRad = 6

DetPreConedX12 = [0.01, 0.01]
DetPreConedX12Mode = 'abs'
DetPreConeds = 0.01
DetPreConedsMode = 'abs'
DetPreConeMarginS = 0.002

DetConeDX = 0.002
DetConeDRY = 0.0025 # 0.0025
DetConeDTheta = np.pi/1024. # 512.
DetConeDZ = 0.0025 # 0.0025
#DetConeNTheta = 25 # 25
#DetConeNZ = 50 # 50

DetConeRefdMax = 0.02

DetEtendMethod = 'quad'
DetEtenddX12 = [0.01, 0.01]
DetEtenddX12Mode = 'rel'
DetEtendepsrel = 1.e-3
DetEtendRatio = 0.02
DetCalcEtendColis = False

DetCalcSAngVectColis = True
DetCalcSAngVectCone = True

DetSynthEpsrel = 1.e-4
DetSynthdX12 = [0.005, 0.005]
DetSynthdX12Mode = 'abs'
DetSynthds = 0.005
DetSynthdsMode = 'abs'
DetSynthMarginS = 0.001

# --- Plotting dictionaries and parameters ------

ApPd = {'c':'k','lw':2,'ls':'solid'}
ApVd = {'color':'r','lw':2,'ls':'solid'}
DetPd = {'c':'k','lw':2,'ls':'solid'}
DetVd = {'color':'r','lw':2,'ls':'solid'}
DetSAngPld = {'cmap':plt.cm.YlOrRd,'lw':0.,'rstride':1,'cstride':1, 'antialiased':False, 'edgecolor':'None'}
DetSangPlContd = {'linewidths':0.}
DetConed = {'edgecolors':'k', 'facecolors':(0.8,0.8,0.8,0.2), 'alpha':0.2, 'linewidths':0., 'linestyles':'-', 'antialiaseds':False}
DetImpd = {'ls':'solid','c':'k','lw':1}

ApLVin = 0.1
DetSAngPlRa = 0.5
DetSAngPldX12 = [0.025,0.025]
DetSAngPldX12Mode = 'rel'
DetSAngPlRatio = 0.01
DetEtendOnLOSNP = 20
DetEtendOnLOSModes = ['trapz']
DetEtendOnLOSLd = {'ls':'solid','c':'k','lw':2}

DetSAngPlot = 'Int'
DetSAngPlotMode = 'scatter'
DetSAngPlotd = {'cmap':plt.cm.YlOrRd}
DetSAngPlotLvl = 20

DetSliceAGdR = 0.005
DetSliceAGdY = 0.005
DetSliceAGdX = 0.01
DetSliceAGdTheta = np.pi/512.
DetSliceAGdZ = 0.005
DetSliceNbd = {'scatter':{'cmap':plt.cm.Greys,'marker':'s','edgecolors':'None','s':10},
        'contour':{'cmap':plt.cm.Greys},
        'contourf':{'cmap':plt.cm.Greys},
        'imshow':{'cmap':plt.cm.Greys}}
DetSliceSAd = {'scatter':{'cmap':plt.cm.YlOrRd,'marker':'s','edgecolors':'None','s':10, 'vmin':0},
        'contour':{'cmap':plt.cm.YlOrRd, 'vmin':0},
        'contourf':{'cmap':plt.cm.YlOrRd, 'vmin':0},
        'imshow':{'cmap':plt.cm.YlOrRd, 'vmin':0}}


DetPolProjNTheta = 50
DetPolProjNZ = 25
DetSAngColis = True

GDetEtendMdA = {'ls':'None','c':'k','lw':2,'marker':'+'}
GDetEtendMdR = {'ls':'None','c':'b','lw':2,'marker':'x'}
GDetEtendMdS = {'ls':'None','c':'g','lw':2,'marker':'o'}
GDetEtendMdP = {'ls':'None','c':'r','lw':2,'marker':'*'}

GDetSigd = {'ls':'solid','c':'k','lw':2,'marker':'+'}


Apertdict = dict(Lax=None, Proj='All', Elt='PV', EltVes='', Leg='', LVIn=ApLVin, Pdict=ApPd, Vdict=ApVd, Vesdict=Vesdict, LegDict=TorLegd, draw=True, Test=True)
#Detdict =


# -------------- Figures ------------------------

def Plot_SAng_Plane_DefAxes(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (10,8,80,'w')
    axPos = [0.05, 0.05, 0.9, 0.9]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,axisbg=axCol,projection='3d')
    ax.set_xlabel(r"X1 (m)")
    ax.set_ylabel(r"X2 (m)")
    ax.set_zlabel(r"$\Omega$ (sr)")
    return ax


def Plot_Etendue_AlongLOS_DefAxes(kMode='rel',a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (14,8,80,'w')
    axPos = [0.06, 0.08, 0.70, 0.86]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
    if kMode.lower()=='rel':
        ax.set_xlabel(r"Rel. length (adim.)")
    else:
        ax.set_xlabel(r"Length (m)")
    ax.set_ylabel(r"Etendue ($sr.m^2$)")
    return ax


def Plot_CrossSlice_SAngNb_DefAxes(VType='Tor', a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (15,8,80,'w')
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    axSAng = f.add_axes([0.05, 0.06, 0.40, 0.85],frameon=True,axisbg=axCol)
    axNb = f.add_axes([0.60, 0.06, 0.40, 0.85],frameon=True,axisbg=axCol)
    if VType=='Tor':
        axSAng.set_xlabel(r"R (m)"),    axNb.set_xlabel(r"R (m)")
    elif VType=='Lin':
        axSAng.set_xlabel(r"Y (m)"),    axNb.set_xlabel(r"Y (m)")
    axSAng.set_ylabel(r"Z (m)"),      axNb.set_ylabel(r"Z (m)")
    axSAng.set_aspect(aspect="equal", adjustable='datalim')
    axNb.set_aspect(aspect="equal", adjustable='datalim')
    return axSAng, axNb


def Plot_HorSlice_SAngNb_DefAxes(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (15,8,80,'w')
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    axSAng = f.add_axes([0.07, 0.12, 0.35, 0.8],frameon=True,axisbg=axCol)
    axNb = f.add_axes([0.55, 0.12, 0.35, 0.8],frameon=True,axisbg=axCol)
    axSAng.set_xlabel(r"X (m)"),    axSAng.set_ylabel(r"Y (m)")
    axNb.set_xlabel(r"X (m)"),      axNb.set_ylabel(r"Y (m)")
    axSAng.set_aspect(aspect="equal", adjustable='datalim')
    axNb.set_aspect(aspect="equal", adjustable='datalim')
    return axSAng, axNb


def Plot_Etendues_GDetect_DefAxes(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (18,8,80,'w')
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes([0.05,0.1,0.85,0.80],frameon=True,axisbg=axCol)
    ax.set_xlabel(r"")
    ax.set_ylabel(r"Etendue (sr.m^2)")
    return ax


def Plot_Sig_GDetect_DefAxes(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (18,8,80,'w')
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes([0.05,0.1,0.85,0.80],frameon=True,axisbg=axCol)
    ax.set_xlabel(r"")
    ax.set_ylabel(r"Signal (W)")
    return ax


#Ldict_mlab_Def = {'color':(0.,0.,0.),'tube_radius':None}
#Mdict_mlab_Def = {'color':(0.,0.,0.),'line_width':1,'mode':'sphere'}
#Dict_3D_mlab_Tor_Def = {'color':(0.8,0.8,0.8),'opacity':0.15,'transparent':False,'scale_factor':0.1}


def Plot_GDetect_Resolution_DefAxes(VType='Tor', a4=False):
    axCol = "w"
    (fW,fH) = (11.69,8.27) if a4 else (16,10)
    f = plt.figure(figsize=(fW,fH),facecolor=axCol)
    ax1 = f.add_axes([0.05, 0.06, 0.32, 0.80], frameon=True, axisbg=axCol)
    ax2 = f.add_axes([0.50, 0.55, 0.47, 0.40], frameon=True, axisbg=axCol)
    ax3 = f.add_axes([0.50, 0.06, 0.47, 0.40], frameon=True, axisbg=axCol)
    X1 = r"R (m)" if VType=='Tor' else r"Y (m)"
    ax1.set_xlabel(X1)
    ax1.set_ylabel(r"Z (m)")
    ax2.set_xlabel(r"size (a.u.)")
    ax2.set_ylabel(r"Signal (mW)")
    ax3.set_xlabel(r"Channels index (from 0)")
    ax3.set_ylabel(r"Signal (mW)")
    ax1.set_aspect(aspect='equal',adjustable='datalim')
    return ax1, ax2, ax3











"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings of ToFu_Mesh
###############################################################################
###############################################################################
"""

# Computing

BF2IntMode = 'Surf'

Mesh1DDefName = 'NoName'
Mesh2DDefName = 'NoName'

L1DRes = 0.001
L1Mode = 'Larger'
L1Tol = 1e-14

L1IntOpSpa = False
L1IntOpSpaFormat = 'dia'


def BF2_DVect_DefR(Points):
    Theta = np.arctan2(Points[1,:],Points[0,:])
    return np.array([np.cos(Theta),np.sin(Theta),np.zeros((Points.shape[1],))])

def BF2_DVect_DefZ(Points):
    return np.array([[0.],[0.],[1.]])*np.ones((1,Points.shape[1]))

def BF2_DVect_DefTor(Points):
    Theta = np.arctan2(Points[1,:],Points[0,:])
    return np.array([-np.sin(Theta),np.cos(Theta),np.zeros((Points.shape[1],))])



# --- Plotting dictionaries and parameters ------


Legpropd = {'size':10}
M2Legd = {'frameon':False,'ncol':1,'bbox_to_anchor':(1.22, 1.12),'loc':2,'borderaxespad':0.5,'prop':Legpropd}

M1Kd = {'c':'b', 'marker':'x', 'markersize':8, 'ls':'None', 'lw':3.}
M1Cd = {'c':'r', 'marker':'o', 'markersize':5, 'ls':'None', 'lw':1.}
M1Resd = {'c':'k', 'ls':'solid', 'lw':2.}

M2Bckd = {'color':(0.9,0.9,0.9), 'marker':'None', 'linestyle':'-', 'linewidth':1.}
M2Mshd = {'color':'k', 'marker':'None', 'linestyle':'-', 'linewidth':0.5}
M2Kd = {'c':'b', 'marker':'x', 'markersize':8, 'ls':'None', 'mew':2.}
M2Cd = {'c':'r', 'marker':'o', 'markersize':6, 'ls':'None', 'mew':0.}
M2Sd = {'cmap':plt.cm.YlOrRd,'edgecolor':None}

BF1Sub = 0.1
BF1SubMode = 'rel'
BF1Fd = {'lw':1,'ls':'solid'}
BF1Totd = {'c':'k','lw':2,'ls':'solid'}

BF2Sub = (0.1,0.1)
BF2SubMode = 'rel'

BF2PlotMode = 'contourf'
BF2PlotSubP = 0.25
BF2PlotSubMode = 'rel'
BF2PlotNC = 25
BF2PlotTotd = {'cmap':plt.cm.YlOrRd,'edgecolor':None}
BF2PlotIndSd = {'edgecolor':(0.8,0.8,0.8,1.),'facecolor':(0.8,0.8,0.8,1.),'linestyle':'solid','linewidth':1}
BF2PlotIndPd = {'c':'g', 'marker':'+', 'markersize':8, 'ls':'None', 'mew':2., 'lw':'none'}



# -------------- Figures ------------------------


def Plot_Mesh1D_DefAxes(a4=False):
    fdpi,axCol = 80,'w'
    (fW,fH) = (11.69,8.27) if a4 else (10,3)
    axPos = [0.04, 0.17, 0.8, 0.7]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
    ax.set_xlabel(r"X (m)")
    return ax

def Plot_Res_Mesh1D_DefAxes(a4=False):
    fdpi,axCol = 80,'w'
    (fW,fH) = (11.69,8.27) if a4 else (10,5)
    axPos = [0.06, 0.17, 0.75, 0.7]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
    ax.set_xlabel(r"X (m)"), ax.set_ylabel(r"$\Delta$ X (m)")
    return ax

def Plot_Mesh2D_DefAxes(VType='Tor', a4=False):
    fdpi,axCol = 80,'w'
    (fW,fH) = (11.69,8.27) if a4 else (8,8)
    axPos = [0.12, 0.08, 0.68, 0.88]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
    Xlab = r"R (m)" if VType=='Tor' else r"Y (m)"
    ax.set_xlabel(Xlab)
    ax.set_ylabel(r"Z (m)")
    ax.set_aspect("equal", adjustable='datalim')
    return ax

def Plot_Res_Mesh2D_DefAxes(a4=False, VType='Tor'):
    fdpi,axCol = 80,'w'
    (fW,fH) = (11.69,8.27) if a4 else (11,11)
    axPos1, axPos2, axPos3, axPoscb = [0.1, 0.07, 0.60, 0.65], [0.75, 0.07, 0.23, 0.65], [0.1, 0.75, 0.60, 0.23], [0.75, 0.75, 0.03, 0.23]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax1 = f.add_axes(axPos1,frameon=True,axisbg=axCol)
    ax2 = f.add_axes(axPos2,frameon=True,axisbg=axCol)
    ax3 = f.add_axes(axPos3,frameon=True,axisbg=axCol)
    axcb = f.add_axes(axPoscb,frameon=True,axisbg=axCol)
    X1str = r"R (m)" if VType=='Tor' else r"Y (m)"
    ax1.set_xlabel(X1str),          ax1.set_ylabel(r"Z (m)")
    ax2.set_xlabel(r"Res. (m)"),    ax2.set_ylabel(r"")
    ax3.set_xlabel(r""),            ax3.set_ylabel(r"Res. (m)")
    ax1.set_aspect("equal", adjustable='datalim')
    return ax1, ax2, ax3, axcb









def Plot_BSpline_DefAxes(Mode):
    if Mode=='1D':
        fW,fH,fdpi,axCol = 12,6,80,'w'
        axPos = [0.08, 0.1, 0.78, 0.8]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
        ax.set_xlabel(r"$X$ ($m$)")
        ax.set_ylabel(r"$Y$ ($a.u.$)")
        return ax
    elif Mode=='2D':
        fW,fH,fdpi,axCol = 10,8,80,'w'
        axPos = [0.1, 0.1, 0.8, 0.8]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
        ax.set_xlabel(r"$R$ ($m$)")
        ax.set_ylabel(r"$Z$ ($m$)")
        ax.axis("equal")
        return ax
    elif Mode=='3D':
        fW,fH,fdpi,axCol = 10,8,80,'w'
        axPos = [0.1, 0.1, 0.8, 0.8]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        ax = f.add_axes(axPos,axisbg=axCol,projection='3d')
        ax.set_xlabel(r"$R$ ($m$)")
        ax.set_ylabel(r"$Z$ ($m$)")
        ax.set_zlabel(r"Emiss. (a.u.)")
        return ax


def Plot_BSplineFit_DefAxes(Mode):
    if Mode=='2D':
        fW,fH,fdpi,axCol = 14,8,80,'w'
        axPos1, axPos2 = [0.1, 0.1, 0.4, 0.8], [0.55, 0.1, 0.4, 0.8]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        ax1 = f.add_axes(axPos1,frameon=True,axisbg=axCol)
        ax2 = f.add_axes(axPos2,frameon=True,axisbg=axCol)
        ax1.set_xlabel(r"$R$ ($m$)"),   ax1.set_ylabel(r"$Z$ ($m$)")
        ax2.set_xlabel(r"$R$ ($m$)")
        ax1.axis("equal"),          ax2.axis("equal")
        return ax1, ax2
    elif Mode=='3D':
        fW,fH,fdpi,axCol = 10,8,80,'w'
        axPos1, axPos2 = [0.1, 0.1, 0.4, 0.8], [0.55, 0.1, 0.4, 0.8]
        f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
        ax1 = f.add_axes(axPos1,axisbg=axCol,projection='3d')
        ax2 = f.add_axes(axPos2,axisbg=axCol,projection='3d')
        ax1.set_xlabel(r"$R$ $(m)$"),           ax1.set_ylabel(r"$Z$ $(m)$")
        ax2.set_xlabel(r"$R$ $(m)$"),           ax2.set_ylabel(r"$Z$ $(m)$")
        ax1.set_zlabel(r"$Emiss.$ $(a.u.)$"),   ax2.set_zlabel(r"$Emiss.$ $(a.u.)$")
        return ax1, ax2



def Plot_BSpline_Deriv_DefAxes(Deg):
    fW,fH,fdpi,axCol = 10,8,80,'w'
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = []
    for ii in range(0,Deg+1):
        ax.append(f.add_subplot(Deg+1,1,ii+1))
        ax[ii].set_xlabel(r"x (a.u.)")
        ax[ii].set_ylabel(r"$d^{"+str(ii)+"}$ (a.u.)")
    return ax


def Plot_BaseFunc2D_BFuncMesh_DefAxes():
    fW,fH,fdpi,axCol = 15,8,80,'w'
    axPos1, axPos2 = [0.05, 0.1, 0.4, 0.8], [0.55, 0.1, 0.4, 0.8]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax1 = f.add_axes(axPos1,frameon=True,axisbg=axCol)
    ax2 = f.add_axes(axPos2,axisbg=axCol,projection='3d')
    ax1.set_xlabel(r"R (m)"), ax2.set_xlabel(r"R (m)")
    ax1.set_ylabel(r"Z (m)"), ax2.set_ylabel(r"Z (m)")
    ax2.set_zlabel(r"Z (a.u.)")
    ax1.axis("equal")
    return ax1, ax2


def Plot_BFunc_SuppMax_PolProj_DefAxes():
    fW,fH,fdpi,axCol = 6,8,80,'w'
    axPos = [0.15, 0.15, 0.6, 0.7]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
    ax.set_xlabel(r"R (m)")
    ax.set_ylabel(r"Z (m)")
    ax.axis("equal")
    return ax

def Plot_BF2_interp_DefAxes():
    fW,fH,fdpi,axCol = 12,8,80,'w'
    axPosV1, axPosC1 = [0.08, 0.05, 0.4, 0.9], [0.56, 0.05, 0.4, 0.9]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    axV1 = f.add_axes(axPosV1,frameon=True,axisbg=axCol)
    axC1 = f.add_axes(axPosC1,frameon=True,axisbg=axCol)
    axV1.set_xlabel(r"R (m)"), axC1.set_xlabel(r"R (m)")
    axV1.set_ylabel(r"Z (m)"), axC1.set_ylabel(r"Z (m)")
    axV1.set_title("Input"), axC1.set_title("Output")
    axV1.axis("equal"), axC1.axis("equal")
    return axV1, axC1












"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings of tofu.Eq
###############################################################################
###############################################################################
"""

# --- Plotting dictionaries and parameters ------

Eq2DPlotDict = {'scatter':{'cmap':plt.cm.YlOrRd, 'marker':'s','edgecolors':'None', 's':10},
        'contour':{'cmap':plt.cm.YlOrRd},
        'contourf':{'cmap':plt.cm.YlOrRd},
        'imshow':{'cmap':plt.cm.YlOrRd}}
Eq2DMagAxDict = {'ls':'None', 'lw':0., 'marker':'+', 'ms':10, 'c':'k', 'mew':2.}
Eq2DSepDict = {'ls':'-', 'lw':1., 'marker':'None', 'c':'k'}

Eq2DPlotRadDict = {'pivot':'tail', 'color':'b', 'units':'xy', 'angles':'xy', 'scale':1., 'scale_units':'xy', 'width':0.003, 'headwidth':3, 'headlength':5}
Eq2DPlotPolDict = {'pivot':'tail', 'color':'r', 'units':'xy', 'angles':'xy', 'scale':1., 'scale_units':'xy', 'width':0.003, 'headwidth':3, 'headlength':5}

Eq2DPlotVsDict = {'ls':'None', 'lw':0., 'marker':'+', 'ms':8, 'color':'k', 'mec':'k'}


# -------------- Figures ------------------------



def Plot_Eq2D_DefAxes(VType='Tor', cbar=False, a4=False):
    fdpi,axCol = 80,'w'
    (fW,fH) = (8.27,11.69) if a4 else (8,10)
    axPos = [0.12, 0.08, 0.68, 0.86]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
    axc = f.add_axes([0.82,0.08,0.05,0.60],frameon=True,axisbg=axCol) if cbar else None
    Xlab = r"R (m)" if VType=='Tor' else r"Y (m)"
    ax.set_xlabel(Xlab)
    ax.set_ylabel(r"Z (m)")
    ax.set_aspect("equal", adjustable='datalim')
    return ax, axc


def Plot_Eq2D_Vs_DefAxes(VType='Tor', a4=False):
    fdpi,axCol = 80,'w'
    (fW,fH) = (11.69,8.27) if a4 else (11,8)
    axPos = [0.07, 0.06, 0.78, 0.87]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes(axPos,frameon=True,axisbg=axCol)
    return ax




def Plot_Eq2D_Inter_DefAxes(VType='Tor', cbar=False, a4=False):
    fdpi,axCol = 80,'w'
    (fW,fH) = (11.69,8.27) if a4 else (16.,11.3)

    f = plt.figure(facecolor="w", figsize=(fW,fH), dpi=fdpi)
    ax1 = f.add_axes([0.05, 0.05, 0.40, 0.90], frameon=True,axisbg=axCol)
    ax2 = f.add_axes([0.55, 0.55, 0.43, 0.40], frameon=True, axisbg=axCol)
    ax3 = f.add_axes([0.55, 0.05, 0.43, 0.40], frameon=True, axisbg=axCol)
    axc = f.add_axes([0.47, 0.05,0.03,0.60],frameon=True,axisbg=axCol) if cbar else None

    Xlab = r"R (m)" if VType=='Tor' else r"Y (m)"
    ax1.set_xlabel(Xlab), ax1.set_ylabel(r"Z (m)")
    ax2.set_xlabel(Xlab)
    ax3.set_xlabel(r"t (s)")
    ax1.set_aspect("equal", adjustable='datalim')

    return {'2DProf':[ax1], '1DProf':[ax2], '1DTime':[ax3], 'Misc':[axc], '1DConst':None, '2DConst':None}











"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings of ToFu_MatComp
###############################################################################
###############################################################################
"""


# ------------ Computing settings ---------------

GMindMSubP = 0.1
GMindMSubPMode = 'Rel'
GMindMSubTheta = 0.02
GMindMSubThetaMode = 'Rel'

GMMatMSubP = 0.1
GMMatMSubPMode = 'Rel'
GMMatMSubTheta = 0.02
GMMatMSubThetaMode = 'Rel'
GMMatepsrel = 1.e-4
GMMatMode = 'trapz'
GMMatDMinInf = 0.0005

GMMatLOSMode = 'quad'
GMMatLOSeps = 1.e-4
GMMatLOSSubP = 0.01
GMMatLOSSubPMode = 'Rel'

GMSigPlotSubP = 0.5
GMSigPlotSubPMode = 'Rel'
GMSigPlotNC = 30


# --- Plotting dictionaries and parameters ------

GMPlotDetSubP = 0.1
GMPlotDetSubPMode = 'Rel'
GMPlotDetKWArgMesh = {'Elt':'M','Mshdict':{'color':(0.9,0.9,0.9), 'marker':'None', 'linestyle':'-', 'linewidth':1.,'zorder':-10}}#'MBKC'
GMPlotDetKWArgTor = {'Elt':'P'}
GMPlotDetKWArgDet = {'Elt':'PC', 'EltApert':'P','EltLOS':'L','EltTor':'P'}


GMPlotBFDetd = {'Elt':'C','Conedict':{'edgecolors':'none','facecolors':(0.8,0.8,0.8,0.2),'linewidths':0.,'zorder':10},'EltLOS':'L','Ldict':{'lw':2,'zorder':10}}

GMPlotBFSubP = 0.05
GMPlotBFSubPMode = 'Rel'
GMPlotBFKWArgLOS = {'Elt':'L'}

GMPlotDetCd = {'cmap':plt.cm.YlOrRd,'edgecolors':'none','linewidths':0.}
GMPlotDetLd = {'lw':2,'ls':'-','c':'k'}
GMPlotDetLOSd = {'lw':2,'ls':'--','c':'k'}

GMSigPlotSd = {'lw':2}
GMSigPlotSLOSd = {'Elt':'L', 'Ldict':{'lw':1}}
GMSigPlotCd = {'cmap':plt.cm.Greys} # or gray_r or Greys or hot_r


# -------------- Figures ------------------------


def Plot_GeomMatrix_Mesh_DefAxes(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (16,10,80,'w')
    axPos1, axPos2, axPos3 = [0.05, 0.07, 0.32, 0.87], [0.48, 0.55, 0.5, 0.4], [0.48, 0.07, 0.5, 0.4]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax1 = f.add_axes(axPos1,frameon=True,axisbg=axCol)
    ax2 = f.add_axes(axPos2,frameon=True,axisbg=axCol)
    ax3 = f.add_axes(axPos3,frameon=True,axisbg=axCol)
    ax1.set_xlabel(r"R (m)"), ax2.set_xlabel(r"Mesh elements index (starts at 0)"), ax3.set_xlabel(r"Basis functions index (starts at 0)")
    ax1.set_ylabel(r"Z (m)"), ax2.set_ylabel(r"Contribution ($W/sr/m^3 x sr.m^3$)"), ax3.set_ylabel(r"Contribution $W/sr/m^3 x sr.m^3$")
    ax1.axis("equal")
    return ax1, ax2, ax3


def Plot_BF2_sum_DefAxes(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (14,10,80,'w')
    axPos1, axPos2 = [0.05, 0.05, 0.85, 0.4], [0.05, 0.55, 0.85, 0.4]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax1 = f.add_axes(axPos1,frameon=True,axisbg=axCol)
    ax2 = f.add_axes(axPos2,frameon=True,axisbg=axCol)
    ax1.set_xlabel(r"Detect index (starts at 0)"), ax2.set_xlabel(r"BFunc index (starts at 0)")
    ax1.set_ylabel(r"Contribution ($sr.m^3$)"), ax2.set_ylabel(r"Contribution ($sr.m^3$)")
    return ax1, ax2


def Plot_SynDiag_DefAxes(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (17,10,80,'w')
    axPos1, axPos2, axPos3, axPos4 = [0.05, 0.60, 0.92, 0.37], [0.08, 0.05, 0.20, 0.50], [0.37, 0.05, 0.26, 0.50], [0.70, 0.05, 0.20, 0.50]
    f = plt.figure(facecolor="w",figsize=(fW,fH),dpi=fdpi)
    ax1 = f.add_axes(axPos1,frameon=True,axisbg=axCol)
    ax2 = f.add_axes(axPos2,frameon=True,axisbg=axCol)
    ax3 = f.add_axes(axPos3,frameon=True,axisbg=axCol)
    ax4 = f.add_axes(axPos4,frameon=True,axisbg=axCol)
    ax1.set_xlabel(r""), ax2.set_xlabel(r"R (m)"), ax3.set_xlabel(r"R (m)"), ax4.set_xlabel(r"R (m)")
    ax1.set_ylabel(r"SXR (W)"), ax2.set_ylabel(r"Z (m)")#, ax3.set_ylabel(r"Z (m)"), ax4.set_ylabel(r"Z (m)")
    ax2.axis("equal"), ax3.axis("equal"), ax4.axis("equal")
    return ax1, ax2, ax3, ax4




"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings of ToFu_Treat
###############################################################################
###############################################################################
"""



def Plot_TreatSig_Def(a4=False, nMax=4):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (22,10,80,'w')
    f = plt.figure(facecolor=axCol,figsize=(fW,fH),dpi=fdpi)
    ax3D = f.add_axes([0.05, 0.55, 0.35, 0.4],frameon=True,axisbg=axCol)
    axtime = f.add_axes([0.05, 0.05, 0.35, 0.4],frameon=True,axisbg=axCol)
    axprof = f.add_axes([0.45, 0.05, 0.53, 0.85],frameon=True,axisbg=axCol)
    LaxTxtChan, LaxTxtTime = [], []
    width, w2 = (0.40-0.13)/nMax, (0.98-0.45)/nMax
    for ii in range(0,nMax):
        LaxTxtChan.append(f.add_axes([0.13+ii*width, 0.45, width, 0.05],frameon=False,axisbg=axCol))
        LaxTxtChan[ii].spines['top'].set_visible(False),    LaxTxtChan[ii].spines['bottom'].set_visible(False)
        LaxTxtChan[ii].spines['right'].set_visible(False),  LaxTxtChan[ii].spines['left'].set_visible(False)
        LaxTxtChan[ii].set_xticks([]),                      LaxTxtChan[ii].set_yticks([])
        LaxTxtChan[ii].set_xlim(0,1),                       LaxTxtChan[ii].set_ylim(0,1)
        LaxTxtTime.append(f.add_axes([0.45+ii*w2, 0.95, w2, 0.04],frameon=False,axisbg=axCol))
        LaxTxtTime[ii].spines['top'].set_visible(False),    LaxTxtTime[ii].spines['bottom'].set_visible(False)
        LaxTxtTime[ii].spines['right'].set_visible(False),  LaxTxtTime[ii].spines['left'].set_visible(False)
        LaxTxtTime[ii].set_xticks([]),                      LaxTxtTime[ii].set_yticks([])
        LaxTxtTime[ii].set_xlim(0,1),                       LaxTxtTime[ii].set_ylim(0,1)
    ax3D.set_xlabel(r"time (s)"), ax3D.set_ylabel(r"Channel index")
    axtime.set_xlabel(r"time (s)"), axtime.set_ylabel(r"SXR Mes. (mW)")
    axprof.set_xlabel(r"Channel index"), axprof.set_ylabel(r"SXR Mes. (mW)")
    axtime.grid(True), axprof.grid(True)
    return ax3D, axtime, axprof, LaxTxtChan, LaxTxtTime

def Plot_Noise_Def(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (20,10,80,'w')
    f = plt.figure(facecolor=axCol,figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes([0.07, 0.08, 0.74, 0.85],frameon=True,axisbg=axCol)
    ax.set_xlabel(r"Phys. (a.u.)"), ax.set_ylabel(r"Noise (a.u.)")
    ax.grid(True)
    return ax


def Plot_FFTChan_Def(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (9,5,80,'w')
    f = plt.figure(facecolor=axCol,figsize=(fW,fH),dpi=fdpi)
    ax = f.add_axes([0.08, 0.1, 0.90, 0.85],frameon=True,axisbg=axCol)
    ax.set_xlabel(r"time (s)")
    ax.set_ylabel(r"Freq. (kHz)")
    ax.grid(True)
    return ax



def Plot_FFTInter_Def(a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (16,11.3,80,'w')
    f = plt.figure(facecolor=axCol,figsize=(fW,fH),dpi=fdpi)
    ax21 = f.add_axes([0.05, 0.55, 0.40, 0.40],frameon=True,axisbg=axCol)
    ax22 = f.add_axes([0.05, 0.05, 0.40, 0.40],frameon=True,axisbg=axCol)
    axt = f.add_axes([0.55, 0.55, 0.40, 0.40],frameon=True,axisbg=axCol)
    axF = f.add_axes([0.55, 0.05, 0.40, 0.40],frameon=True,axisbg=axCol)
    ax21.set_ylabel(r"Freq. (kHz)"), ax21.set_title(r"Pow. spectrum norm. to max")
    ax22.set_xlabel(r"time (s)"),   ax22.set_ylabel(r"Freq. (kHz)"), ax22.set_title(r"Pow. spectrum norm. to instantaneous max")
    axt.set_xlabel(r"time (s)"),    axt.set_ylabel(r"Harm. magnitude$^2$ (a.u.)")
    axF.set_xlabel(r"Freq. (kHz)"),    axF.set_ylabel(r"Harm. magnitude$^2$ (a.u.)")
    ax21.grid(True), ax22.grid(True), axt.grid(True), axF.grid(True)
    return ax21, ax22, axt, axF






"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings of ToFu_Inv
###############################################################################
###############################################################################
"""




# Probability law for augmented tikho : p(x) = x^(a-1)*exp(-bx)
# If normalised :
#       Mean [x]        a / b
#       Variance <x>    a / b^2
# If not normalised :
#       Mean [x]        a! / (b^(a+1))
#       Variance <x>    a!/b^(2a+2) * ( (a+1)b^a - a! )
#       [x] = k     =>  b = (a!/k)^(1/(a+1))
#       if [x]=1    =>  <x> = (a+1)* (a!)^(1/(a+1)) - 1


ConvCrit = 1.e-6
chi2Tol = 0.05
chi2Obj = 1.
mu0 = 1000.


AugTikho_a0 = 10                                                    # (Regul. parameter, larger a => larger variance)
AugTikho_b0 = np.math.factorial(AugTikho_a0)**(1/(AugTikho_a0+1))   # To have [x] = 1
AugTikho_a1 = 2                                                     # (Noise), a as small as possible for small variance
AugTikho_b1 = np.math.factorial(AugTikho_a1)**(1/(AugTikho_a1+1))   # To have [x] = 1
AugTikho_d = 0.95                                                   # Exponent for rescaling of a0bis in V2, typically in [1/3 ; 1/2], but real limits are 0 < d < 1 (or 2 ?)

AugTkLsmrAtol = 1.e-8
AugTkLsmrBtol = 1.e-8
AugTkLsmrConlim = 1e8
AugTkLsmrMaxiter = None


SolInvParDef = {'Dt':None,'mu0':mu0,'SolMethod':'InvLin_AugTikho_V1','Deriv':'D2N2','IntMode':'Vol','Cond':None,'ConvCrit':ConvCrit,'Sparse':True,'SpType':'csr', 'Sep':{'In':True,'NLim':3}, 'Pos':True, 'KWARGS': {'a0':AugTikho_a0, 'b0':AugTikho_b0, 'a1':AugTikho_a1, 'b1':AugTikho_b1, 'd':AugTikho_d, 'ConvReg':True, 'FixedNb':True}, 'timeit':False, 'Verb':False,'VerbNb':None,'Method':'Poly','Deg':1,'Group':True,'plot':False,'LNames':None,'Com':''}


# --- Plotting dictionaries and parameters ------

InvPlotSubP = 0.01
InvPlotSubMode = 'abs'

InvAnimIntervalms = 100
InvAnimBlit = True
InvAnimRepeat = True
InvAnimRepeatDelay = 500
InvAnimTimeScale = 1.e2

InvPlotF = 'imshow'
InvLvls = 30
Invdict = {'cmap':plt.cm.YlOrRd,'edgecolor':None}
Tempd = {'ls':'-','c':'k'}
Retrod = {'ls':'-','c':'b'}
InvSXRd = {'ls':'-','c':'k','lw':1.}
InvSigmad = {'facecolor':(0.8,0.8,0.8,0.7),'lw':0.}
"""
"""
# -------------- Figures ------------------------

def Plot_Inv_Anim_DefAxes(SXR=True, TMat=True, Chi2N=True, Mu=True, R=True, Nit=True):
    axCol = 'w'
    if not any([SXR,TMat,Chi2N,Mu,R,Nit]):
        fW,fH = 8,10
        f = plt.figure(figsize=(fW,fH),facecolor=axCol)
        axInvPos = [0.10, 0.10, 0.75, 0.80]
        axcPos = [0.87, 0.10, 0.10, 0.70]
        axTMat = None
        tempPos = []

    elif SXR and not any([TMat,Chi2N,Mu,R,Nit]):
        fW,fH = 8,12
        f = plt.figure(figsize=(fW,fH),facecolor=axCol)
        axInvPos = [0.10, 0.30, 0.70, 0.65]
        axcPos = [0.82, 0.30, 0.10, 0.55]
        tempPos = [[0.10, 0.05, 0.85, 0.20]]
        tempylab = [r"SXR (mW)"]
        axTMat = None

    elif SXR and TMat and not any([Chi2N,Mu,R,Nit]):
        fW,fH = 16,10
        f = plt.figure(figsize=(fW,fH),facecolor=axCol)
        axInvPos = [0.05, 0.06, 0.32, 0.80]
        axcPos = [0.38, 0.06, 0.03, 0.70]
        tempPos = [[0.50, 0.06, 0.47, 0.40]]
        tempylab = [r"SXR (mW)"]
        TMatPos = [0.50, 0.55, 0.47, 0.40]
        TMatylab = r"SXR (mW)"

    elif not SXR and any([Chi2N,Mu,R,Nit]):
        fW,fH = 16,10
        f = plt.figure(figsize=(fW,fH),facecolor=axCol)
        axInvPos = [0.05, 0.06, 0.35, 0.80]
        axcPos = [0.41, 0.06, 0.04, 0.70]
        tempylab = [r"Nb. iterations", r"$\chi^2_N$", r"Reg. param. (a.u.)", r"Obj. func. (a.u.)"]
        tempPos = [[0.50, 0.75, 0.47, 0.21], [0.50, 0.52, 0.47, 0.21], [0.50, 0.29, 0.47, 0.21], [0.50, 0.06, 0.47, 0.21]]
        temps = [Nit,Chi2N,Mu,R]
        for ii in range(0,len(temps)):
            if not temps[ii]:
                del tempPos[ii]
                del tempylab[ii]
        axTMat = None

    else:
        fW,fH = 18,12
        f = plt.figure(figsize=(fW,fH),facecolor=axCol)
        axInvPos = [0.05, 0.05, 0.32, 0.80]
        axcPos = [0.38, 0.05, 0.03, 0.70]
        tempylab = [r"SXR (mW)", r"Nb. iterations", r"$\mathbf{\chi^2_N}$", r"Reg. param. (a.u.)", r"Obj. func. (a.u.)"]
        tempPos = [[0.50, 0.61, 0.47, 0.17], [0.50, 0.47, 0.47, 0.12], [0.50, 0.33, 0.47, 0.12], [0.50, 0.19, 0.47, 0.12], [0.50, 0.05, 0.47, 0.12]]
        temps = [SXR,Nit,Chi2N,Mu,R]
        for ii in range(0,len(temps)):
            if not temps[ii]:
                del tempPos[ii]
                del tempylab[ii]
        if TMat:
            TMatPos = [0.50, 0.81, 0.47, 0.17]
            TMatylab = r"SXR (mW)"
        else:
            axTMat = None


    if TMat:
        axTMat = f.add_axes(TMatPos, frameon=True, axisbg=axCol)
        axTMat.set_ylabel(TMatylab, fontsize=12, fontweight='bold')

    Laxtemp = []
    if len(tempPos)>0:
        ypos = np.array([pp[1] for pp in tempPos])
        indmin = np.argmin(ypos)
        for ii in range(0,len(tempPos)):
            Laxtemp.append(f.add_axes(tempPos[ii], frameon=True, axisbg=axCol))
            Laxtemp[-1].set_ylabel(tempylab[ii], fontsize=12, fontweight='bold')
            if not ii == indmin:
                Laxtemp[-1].set_xticklabels([])
        Laxtemp[np.argmin(ypos)].set_xlabel(r"t (s)", fontsize=12, fontweight='bold')

    axInv = f.add_axes(axInvPos, frameon=True, axisbg=axCol)
    axInv.set_xlabel(r"R (m)", fontsize=12, fontweight='bold')
    axInv.set_ylabel(r"Z (m)", fontsize=12, fontweight='bold')
    axInv.axis('equal')
    axc = f.add_axes(axcPos, frameon=True, axisbg=axCol)

    return axInv, axTMat, Laxtemp, axc



def Plot_Inv_FFTPow_DefAxes(NPts=6, a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (18,12,80,'w')
    f = plt.figure(figsize=(fW,fH),facecolor=axCol,dpi=fdpi)
    axInv = f.add_axes([0.05, 0.05, 0.32, 0.80], frameon=True, axisbg=axCol)
    axc = f.add_axes([0.38, 0.05, 0.03, 0.70], frameon=True, axisbg=axCol)
    axRetro = f.add_axes([0.50, 0.81, 0.47, 0.17], frameon=True, axisbg=axCol)
    axSXR = f.add_axes([0.50, 0.61, 0.47, 0.17], frameon=True, axisbg=axCol)
    axInv.set_xlabel(r"R (m)", fontsize=12, fontweight='bold')
    axInv.set_ylabel(r"Z (m)", fontsize=12, fontweight='bold')
    axInv.axis('equal')
    axRetro.set_xlabel(r"Channels"), axRetro.set_ylabel(r"SXR (mW)")
    axSXR.set_ylabel(r"SXR (mW)")

    Lax = []
    if NPts <= 4:
        N2 = 1
        Hei =  (0.58-0.05)/NPts - (NPts-1)*Dis
        for ii in range(0,NPts):
            Lax.append(f.add_axes([0.50, 0.58-(ii+1)*Hei-ii*Dis, 0.47, Hei], frameon=True, axisbg=axCol))
            Lax[-1].set_ylabel(r"Freq. (kHz)", fontsize=12, fontweight='bold')
            Lax[-1].grid(True)
        Lax[-1].set_xlabel(r"time (s)", fontsize=12, fontweight='bold')
    else:
        N2, Dis = int(np.ceil(NPts/2.)), 0.03
        Hei = (0.58-0.05-(N2-1)*Dis)/N2
        LaxPos = [[[0.50, 0.58-(ii+1)*Hei-ii*Dis, 0.22, Hei],[0.75, 0.58-(ii+1)*Hei-ii*Dis, 0.22, Hei]] for ii in range(0,N2)]
        #LaxPos = list(itt.chain._from_iterable(LaxPos))
        for ii in range(0,N2):
            Lax.append(f.add_axes(LaxPos[ii][0], frameon=True, axisbg=axCol))
            Lax[-1].set_ylabel(r"Freq. (kHz)", fontsize=12, fontweight='bold')
            Lax[-1].grid(True)
            Lax.append(f.add_axes(LaxPos[ii][1], frameon=True, axisbg=axCol))
            Lax[-1].grid(True)
            Lax[-1].set_xticklabels([]), Lax[-1].set_yticklabels([])
            Lax[-2].set_xticklabels([])
        Lax[-2].set_xlabel(r"time (s)", fontsize=12, fontweight='bold')
        Lax[-1].set_xlabel(r"time (s)", fontsize=12, fontweight='bold')
    return axInv, axc, axRetro, axSXR, Lax



"""
###############################################################################
###############################################################################
###############################################################################
                Defaults settings of ToFu_PostTreat
###############################################################################
###############################################################################
"""



# --- Plotting dictionaries and parameters ------

InvPlotSubP = 0.01
InvPlotSubMode = 'abs'
InvPlotF = 'contour' #'imshow'
InvLvls = 30
Invdict = {'cmap':plt.cm.YlOrRd,'edgecolor':None}
Tempd = {'ls':'-','c':'k'}
Retrod = {'ls':'-'}
vlined = {'c':'k','ls':'--','lw':1.}
InvSXRd = {'ls':'-','c':'k','lw':1.}
InvSigmad = {'facecolor':(0.8,0.8,0.8,0.7),'lw':0.}
InvPlot_LPath = [((1.70,0.),(1.,0.)), ((1.70,0.),(0.,1.)), ((1.70,0.20),(1.,0.)), ((1.70,-0.20),(1.,0.)), ((1.55,0.),(0.,1.)), ((1.85,0.),(0.,1.))]
InvPlot_dl = 0.0025
InvPlot_LCol = ['b','r','g','m','y','c']

TFPT_prop = {'Deriv':0, 'indt0':0, 't0':None, 'DVect':BF2_DVect_DefR, 'SubP':InvPlotSubP, 'SubMode':InvPlotSubMode, 'InvPlotFunc':InvPlotF, 'InvLvls':InvLvls, 'Invd':Invdict,
        'vlined':vlined, 'SXRd':InvSXRd, 'Sigmad':InvSigmad, 'Tempd':Tempd, 'Retrod':Retrod, 'VMinMax':[None,None], 'Com':'', 'Norm':False, 'a4':False}

TFPT_propbasic = {}
TFPT_proptechnical = {}
TFPT_propprofiles= {'LPath':InvPlot_LPath, 'dl':InvPlot_dl}

TFPT_propbasic.update(TFPT_prop)
TFPT_proptechnical.update(TFPT_prop)
TFPT_propprofiles.update(TFPT_prop)



TFPT_Lprop = {'basic':TFPT_propbasic, 'technical':TFPT_proptechnical, 'profiles':TFPT_propprofiles, 'sawtooth':None}


InvAnimIntervalms = 100
InvAnimBlit = True
InvAnimRepeat = True
InvAnimRepeatDelay = 500
InvAnimTimeScale = 1.e2





# -------------- Figures ------------------------


def Plot_Inv_Basic_DefAxes(a4=False, dpi=80):
    axCol = 'w'
    (fW,fH) = (16,10) if not a4 else (11.69,8.27)
    f = plt.figure(figsize=(fW,fH),facecolor=axCol,dpi=dpi)
    axInv   =   f.add_axes([0.05, 0.06, 0.32, 0.80], frameon=True, axisbg=axCol)
    axc     =   f.add_axes([0.38, 0.06, 0.03, 0.70], frameon=True, axisbg=axCol)
    axTMat  =   f.add_axes([0.50, 0.55, 0.47, 0.40], frameon=True, axisbg=axCol)
    axSig   =   f.add_axes([0.50, 0.06, 0.47, 0.40], frameon=True, axisbg=axCol)
    axTxt   =   f.add_axes([0.05, 0.86, 0.32, 0.04], frameon=False, axisbg='none')

    axInv.set_xlabel(r"R (m)", fontsize=12, fontweight='bold')
    axInv.set_ylabel(r"Z (m)", fontsize=12, fontweight='bold')
    axTMat.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')
    axSig.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')
    axSig.set_xlabel(r"t (s)", fontsize=12, fontweight='bold')
    axTxt.set_xticks([]), axTxt.set_yticks([])
    axTxt.set_xticklabels([]), axTxt.set_yticklabels([])
    axc.set_title(r"$\epsilon$ (W/m^3)", size=11)

    axInv.axis('equal')
    return axInv, axTMat, axSig, axc, axTxt


def Plot_Inv_Technical_DefAxes(a4=False, dpi=80):
    axCol = 'w'
    (fW,fH) = (18,12) if not a4 else (11.69,8.27)
    f = plt.figure(figsize=(fW,fH),facecolor=axCol,dpi=dpi)
    axInv   =   f.add_axes([0.05, 0.05, 0.32, 0.80], frameon=True, axisbg=axCol)
    axc     =   f.add_axes([0.38, 0.05, 0.03, 0.70], frameon=True, axisbg=axCol)
    axTMat  =   f.add_axes([0.50, 0.81, 0.47, 0.17], frameon=True, axisbg=axCol)
    axSig   =   f.add_axes([0.50, 0.61, 0.47, 0.17], frameon=True, axisbg=axCol)
    axNit   =   f.add_axes([0.50, 0.47, 0.47, 0.12], frameon=True, axisbg=axCol)
    axChi2N =   f.add_axes([0.50, 0.33, 0.47, 0.12], frameon=True, axisbg=axCol)
    axMu    =   f.add_axes([0.50, 0.19, 0.47, 0.12], frameon=True, axisbg=axCol)
    axR     =   f.add_axes([0.50, 0.05, 0.47, 0.12], frameon=True, axisbg=axCol)
    axTxt   =   f.add_axes([0.05, 0.86, 0.32, 0.04], frameon=False, axisbg='none')

    axInv.set_xlabel(r"R (m)", fontsize=12, fontweight='bold'),                 axInv.set_ylabel(r"Z (m)", fontsize=12, fontweight='bold')
    axTMat.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold'),          axSig.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')
    axNit.set_ylabel(r"Nb. iterations", fontsize=12, fontweight='bold')
    axChi2N.set_ylabel(r"$\mathbf{\chi^2_N}$", fontsize=12, fontweight='bold')
    axMu.set_ylabel(r"Reg. param. (a.u.)", fontsize=12, fontweight='bold')
    axR.set_xlabel(r"t (s)", fontsize=12, fontweight='bold'),                   axR.set_ylabel(r"Obj. func. (a.u.)", fontsize=12, fontweight='bold')
    axTxt.set_xticks([]), axTxt.set_yticks([]),                                 axTxt.set_xticklabels([]), axTxt.set_yticklabels([])
    axc.set_title(r"$\epsilon$ (W/m^3)", size=11)

    axInv.axis('equal')
    return axInv, axTMat, [axSig,axNit,axChi2N,axMu,axR], axc, axTxt


def Plot_Inv_Profiles_DefAxes(NL=4, a4=False, dpi=80):
    axCol = 'w'
    (fW,fH) = (18,12) if not a4 else (11.69,8.27)
    f = plt.figure(figsize=(fW,fH),facecolor=axCol,dpi=dpi)
    axInv   =   f.add_axes([0.05, 0.05, 0.32, 0.80], frameon=True, axisbg=axCol)
    axc     =   f.add_axes([0.38, 0.05, 0.03, 0.70], frameon=True, axisbg=axCol)
    axTMat  =   f.add_axes([0.50, 0.83, 0.47, 0.15], frameon=True, axisbg=axCol)
    axSig   =   f.add_axes([0.50, 0.65, 0.47, 0.15], frameon=True, axisbg=axCol)
    axInv.set_xlabel(r"R (m)", fontsize=12, fontweight='bold'),                 axInv.set_ylabel(r"Z (m)", fontsize=12, fontweight='bold')
    axTMat.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold'),          axSig.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')

    LaxP, RefUp, RefDo, Dz = [], 0.60, 0.05, 0.05
    nl = (NL+1)/2
    H = (RefUp-RefDo - (nl-1)*Dz)/nl
    for ii in range(0,NL):
        cc = ii/nl
        il = ii-cc*nl
        ax = f.add_axes([0.50 + cc*0.25, RefUp-(il+1)*H-il*Dz, 0.22, H], frameon=True, axisbg=axCol)
        LaxP.append(ax)
        if il == nl-1:
            LaxP[ii].set_xlabel(r"Length (m)", fontsize=12, fontweight='bold')
        if cc==0:
            LaxP[ii].set_ylabel(r"$\mathbf{\epsilon^{\eta}}$ (W/m^3)", fontsize=12, fontweight='bold')
    axTxt   =   f.add_axes([0.05, 0.86, 0.32, 0.04], frameon=False, axisbg='none')
    axTxt.set_xticks([]), axTxt.set_yticks([]),                                 axTxt.set_xticklabels([]), axTxt.set_yticklabels([])
    axc.set_title(r"$\epsilon$ (W/m^3)", size=11)

    axInv.set_aspect(aspect='equal',adjustable='datalim')
    return axInv, axTMat, axSig, LaxP, axc, axTxt


def Plot_Inv_Compare_Basic_DefAxes(N=2, a4=False, dpi=80):

    MR, ML, DX = 0.08, 0.06, 0.04
    W = (1.-(MR+ML)-(N-1)*DX)/N

    axCol = 'w'
    (fW,fH) = (16,10) if not a4 else (11.69,8.27)
    f = plt.figure(figsize=(fW,fH),facecolor=axCol,dpi=dpi)

    axTMat  =   f.add_axes([0.55, 0.06, 0.44, 0.34], frameon=True, axisbg=axCol)
    axSig   =   f.add_axes([0.06, 0.06, 0.44, 0.34], frameon=True, axisbg=axCol)
    axc     =   f.add_axes([1.-MR+0.01, 0.46, 0.02, 0.44], frameon=True, axisbg=axCol)

    LaxInv, LaxTxt = [], []
    for ii in range(0,N):
        LaxInv.append(f.add_axes([ML+ii*(DX+W), 0.46, W, 0.44], frameon=True, axisbg=axCol))
        LaxTxt.append(f.add_axes([ML+ii*(DX+W), 0.90, W, 0.03], frameon=False, axisbg='none'))
        LaxInv[ii].set_xlabel(r"R (m)", fontsize=12, fontweight='bold')
        if ii==0:
            LaxInv[ii].set_ylabel(r"Z (m)", fontsize=12, fontweight='bold')
        LaxTxt[ii].set_xticks([]),      LaxTxt[ii].set_yticks([])
        LaxTxt[ii].set_xticklabels([]), LaxTxt[ii].set_yticklabels([])
        LaxInv[ii].axis('equal')

    #axTMat.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')
    axc.set_title(r"$\epsilon$ (W/m^3)", size=11)
    axSig.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')
    axSig.set_xlabel(r"t (s)", fontsize=12, fontweight='bold')
    axTMat.set_xlabel(r"Chan.", fontsize=12, fontweight='bold')

    return LaxInv, axTMat, axSig, axc, LaxTxt


def Plot_Inv_Compare_Technical_DefAxes(N=2, a4=False, dpi=80):

    MR, ML, DX = 0.06, 0.05, 0.04
    W = (1.-(MR+ML)-(N-1)*DX)/N

    axCol = 'w'
    (fW,fH) = (18,12) if not a4 else (11.69,8.27)
    f = plt.figure(figsize=(fW,fH),facecolor=axCol,dpi=dpi)

    axTMat  =   f.add_axes([0.05, 0.26, 0.28, 0.17], frameon=True, axisbg=axCol)
    axSig   =   f.add_axes([0.05, 0.05, 0.28, 0.17], frameon=True, axisbg=axCol)
    axNit   =   f.add_axes([0.38, 0.05, 0.28, 0.18], frameon=True, axisbg=axCol)
    axChi2N =   f.add_axes([0.38, 0.25, 0.28, 0.18], frameon=True, axisbg=axCol)
    axMu    =   f.add_axes([0.70, 0.05, 0.28, 0.18], frameon=True, axisbg=axCol)
    axR     =   f.add_axes([0.70, 0.25, 0.28, 0.18], frameon=True, axisbg=axCol)
    axc     =   f.add_axes([1.-MR+0.01, 0.46, 0.02, 0.44], frameon=True, axisbg=axCol)

    LaxInv, LaxTxt = [], []
    for ii in range(0,N):
        LaxInv.append(f.add_axes([ML+ii*(DX+W), 0.48, W, 0.42], frameon=True, axisbg=axCol))
        LaxTxt.append(f.add_axes([ML+ii*(DX+W), 0.90, W, 0.03], frameon=False, axisbg='none'))
        LaxInv[ii].set_xlabel(r"R (m)", fontsize=12, fontweight='bold')
        if ii==0:
            LaxInv[ii].set_ylabel(r"Z (m)", fontsize=12, fontweight='bold')
        LaxTxt[ii].set_xticks([]),      LaxTxt[ii].set_yticks([])
        LaxTxt[ii].set_xticklabels([]), LaxTxt[ii].set_yticklabels([])
        LaxInv[ii].axis('equal')

    #axTMat.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')
    axc.set_title(r"$\epsilon$ (W/m^3)", size=11)
    axSig.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')
    axSig.set_xlabel(r"t (s)", fontsize=12, fontweight='bold')
    axTMat.set_xlabel(r"Chan.", fontsize=12, fontweight='bold')
    axNit.set_xlabel(r"t (s)"),  axNit.set_ylabel(r"$N_{it}$")
    axChi2N.set_ylabel(r"$\chi^2_N$")
    axMu.set_xlabel(r"t (s)"),  axMu.set_ylabel(r"Reg. param. (a.u.)")
    axR.set_ylabel(r"Obj. func. (a.u.)")

    return LaxInv, axTMat, [axSig,axNit,axChi2N,axMu,axR], axc, LaxTxt


def Plot_Inv_Compare_Profiles_DefAxes(N=2, NL=4, a4=False, dpi=80):

    MR, ML, DX = 0.06, 0.05, 0.04
    W = (1.-(MR+ML)-(N-1)*DX)/N

    axCol = 'w'
    (fW,fH) = (18,12) if not a4 else (11.69,8.27)
    f = plt.figure(figsize=(fW,fH),facecolor=axCol,dpi=dpi)

    axTMat  =   f.add_axes([0.05, 0.29, 0.28, 0.20], frameon=True, axisbg=axCol)
    axSig   =   f.add_axes([0.05, 0.05, 0.28, 0.20], frameon=True, axisbg=axCol)
    axc     =   f.add_axes([1.-MR+0.01, 0.52, 0.02, 0.38], frameon=True, axisbg=axCol)

    LaxInv, LaxTxt = [], []
    for ii in range(0,N):
        LaxInv.append(f.add_axes([ML+ii*(DX+W), 0.54, W, 0.38], frameon=True, axisbg=axCol))
        LaxTxt.append(f.add_axes([ML+ii*(DX+W), 0.92, W, 0.03], frameon=False, axisbg='none'))
        LaxInv[ii].set_xlabel(r"R (m)", fontsize=12, fontweight='bold')
        if ii==0:
            LaxInv[ii].set_ylabel(r"Z (m)", fontsize=12, fontweight='bold')
        LaxTxt[ii].set_xticks([]),      LaxTxt[ii].set_yticks([])
        LaxTxt[ii].set_xticklabels([]), LaxTxt[ii].set_yticklabels([])
        LaxInv[ii].axis('equal')

    LaxP, LaxPbis, RefUp, RefDo, Dz = [], [], 0.49, 0.05, 0.04
    nl = (NL+1)/2
    H = (RefUp-RefDo - (nl-1)*Dz)/nl
    for ii in range(0,NL):
        cc = ii/nl
        il = ii-cc*nl
        ax = f.add_axes([0.38 + cc*0.32, RefUp-(il+1)*H-il*Dz, 0.28, H], frameon=True, axisbg=axCol)
        LaxP.append(ax)
        #LaxPbis.append(ax.twiny())
        if il == nl-1:
            LaxP[ii].set_xlabel(r"Length (m)", fontsize=12, fontweight='bold')
        if cc==0:
            LaxP[ii].set_ylabel(r"$\mathbf{\epsilon^{\eta}}$ (W/m^3)", fontsize=12, fontweight='bold')

    #axTMat.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')
    axc.set_title(r"$\epsilon$ (W/m^3)", size=11)
    axSig.set_ylabel(r"Signal (mW)", fontsize=12, fontweight='bold')
    axSig.set_xlabel(r"t (s)", fontsize=12, fontweight='bold')
    axTMat.set_xlabel(r"Chan.", fontsize=12, fontweight='bold')

    return LaxInv, axTMat, axSig, LaxP, LaxPbis, axc, LaxTxt


















def Plot_Inv_FFTPow_DefAxes(NPts=6, a4=False):
    (fW,fH,fdpi,axCol) = (11.69,8.27,80,'w') if a4 else (18,12,80,'w')
    f = plt.figure(figsize=(fW,fH),facecolor=axCol,dpi=fdpi)
    axInv = f.add_axes([0.05, 0.05, 0.32, 0.80], frameon=True, axisbg=axCol)
    axc = f.add_axes([0.38, 0.05, 0.03, 0.70], frameon=True, axisbg=axCol)
    axRetro = f.add_axes([0.50, 0.81, 0.47, 0.17], frameon=True, axisbg=axCol)
    axSXR = f.add_axes([0.50, 0.61, 0.47, 0.17], frameon=True, axisbg=axCol)
    axInv.set_xlabel(r"R (m)", fontsize=12, fontweight='bold')
    axInv.set_ylabel(r"Z (m)", fontsize=12, fontweight='bold')
    axInv.axis('equal')
    axRetro.set_xlabel(r"Channels"), axRetro.set_ylabel(r"SXR (mW)")
    axSXR.set_ylabel(r"SXR (mW)")

    Lax = []
    if NPts <= 4:
        N2 = 1
        Hei =  (0.58-0.05)/NPts - (NPts-1)*Dis
        for ii in range(0,NPts):
            Lax.append(f.add_axes([0.50, 0.58-(ii+1)*Hei-ii*Dis, 0.47, Hei], frameon=True, axisbg=axCol))
            Lax[-1].set_ylabel(r"Freq. (kHz)", fontsize=12, fontweight='bold')
            Lax[-1].grid(True)
        Lax[-1].set_xlabel(r"time (s)", fontsize=12, fontweight='bold')
    else:
        N2, Dis = int(np.ceil(NPts/2.)), 0.03
        Hei = (0.58-0.05-(N2-1)*Dis)/N2
        LaxPos = [[[0.50, 0.58-(ii+1)*Hei-ii*Dis, 0.22, Hei],[0.75, 0.58-(ii+1)*Hei-ii*Dis, 0.22, Hei]] for ii in range(0,N2)]
        #LaxPos = list(itt.chain._from_iterable(LaxPos))
        for ii in range(0,N2):
            Lax.append(f.add_axes(LaxPos[ii][0], frameon=True, axisbg=axCol))
            Lax[-1].set_ylabel(r"Freq. (kHz)", fontsize=12, fontweight='bold')
            Lax[-1].grid(True)
            Lax.append(f.add_axes(LaxPos[ii][1], frameon=True, axisbg=axCol))
            Lax[-1].grid(True)
            Lax[-1].set_xticklabels([]), Lax[-1].set_yticklabels([])
            Lax[-2].set_xticklabels([])
        Lax[-2].set_xlabel(r"time (s)", fontsize=12, fontweight='bold')
        Lax[-1].set_xlabel(r"time (s)", fontsize=12, fontweight='bold')
    return axInv, axc, axRetro, axSXR, Lax








# Built-in
import os
import shutil
import requests
import warnings

# Common
import numpy as np


__all__ = ['search_online', 'search_online_by_wavelengthA',
           'download', 'download_all', 'clean_downloads']


# Check whether a local .tofu/ repo exists
_URL = 'https://open.adas.ac.uk'
_URL_SEARCH = _URL + '/freeform?searchstring='
_URL_SEARCH_WAVL = _URL + '/wavelength?'
_URL_ADF15 = _URL + '/adf15'
_URL_DOWNLOAD = _URL + '/download'

_INCLUDE_PARTIAL = True


# #############################################################################
#                           Utility functions
# #############################################################################


def _get_PATH_LOCAL():
    pfe = os.path.join(os.path.expanduser('~'), '.tofu', 'openadas2tofu')
    if os.path.isdir(pfe):
        return pfe
    else:
        return None


def _getcharray(ar, col=None, sep='  ', line='-', just='l',
                verb=True, returnas=str):
    """ Format and return char array (for pretty printing) """
    c0 = ar is None or len(ar) == 0
    if c0:
        return ''
    ar = np.array(ar, dtype='U')

    if ar.ndim == 1:
        ar = ar.reshape((1, ar.size))

    # Get just len
    nn = np.char.str_len(ar).max(axis=0)
    if col is not None:
        if len(col) not in ar.shape:
            msg = ("len(col) should be in np.array(ar, dtype='U').shape:\n"
                   + "\t- len(col) = {}\n".format(len(col))
                   + "\t- ar.shape = {}".format(ar.shape))
            raise Exception(msg)
        if len(col) != ar.shape[1]:
            ar = ar.T
            nn = np.char.str_len(ar).max(axis=0)
        nn = np.fmax(nn, [len(cc) for cc in col])

    # Apply to array
    fjust = np.char.ljust if just == 'l' else np.char.rjust
    out = np.array([sep.join(v) for v in fjust(ar, nn)])

    # Apply to col
    if col is not None:
        arcol = np.array([col, [line*n for n in nn]], dtype='U')
        arcol = np.array([sep.join(v) for v in fjust(arcol, nn)])
        out = np.append(arcol, out)

    if verb is True:
        print('\n'.join(out))
    if returnas is str:
        return '\n'.join(out)
    elif returnas is np.ndarray:
        return ar


# #############################################################################
#                          online search
# #############################################################################


class FileAlreayExistsException(Exception):
    pass


def search_online(searchstr=None, returnas=None,
                  include_partial=None, verb=None):
    """ Perform an online freeform search on https://open.adas.ac.uk

    Pass searchstr to the online freeform search
    Prints the results (if verb=True)
    Optionally return the result as:
        - a char array (returnas = np.ndarray)
        - a formatted str (returnas = str)

    example
    -------
        >>> import tofu as tf
        >>> tf.openadas2tofu.search_online('ar+16 ADF15')
    """

    # Check input
    if returnas is None:
        returnas = False
    if verb is None:
        verb = True
    if include_partial is None:
        include_partial = _INCLUDE_PARTIAL
    if searchstr is None:
        searchstr = ''
    searchurl = '+'.join([requests.utils.quote(kk)
                          for kk in searchstr.split(' ')])

    total_url = '{}{}&{}'.format(_URL_SEARCH, searchurl, 'searching=1')
    resp = requests.get(total_url)

    # Extract response from html
    out = resp.text.split('\n')
    flag0 = '<table summary="Freeform search results">'
    flag1 = '</table>'
    ind0 = [ii for ii, vv in enumerate(out) if flag0 in vv]
    ind1 = [ii for ii, vv in enumerate(out) if flag1 in vv]
    if len(ind0) != 1 or len(ind1) == 0:
        msg = ("Format of html response seems to have changed!\n"
               + "Cannot find flags:\n"
               + "\t- {}\n".format(flag0)
               + "\t- {}\n".format(flag1)
               + "in requests.get({}).text".format(total_url))
        raise Exception(msg)
    ind1 = np.min([ii for ii in ind1 if ii > ind0[0]])
    out = out[ind0[0] + 1:ind1-1]
    nresults = len(out) - 1

    # Get columns
    heads = [str.replace(kk.replace('<tr>', '').replace('<th>', ''),
                         '</th>', '').replace('</tr>', '')
             for kk in out[0].split('</th><th>')]
    nhead = len(heads)

    # Get results
    lout = []
    for ii in range(0, nresults):
        if 'Partial results are listed below' in out[ii+1]:
            if include_partial is False:
                break
            lout.append(['-', '-', '----- (partial results) -----', '-'])
            ind = out[ii+1].index('below</th></tr>') + len('below</th></tr>')
            out[ii+1] = out[ii+1][ind:]
        lstri = out[ii+1].split('</td><td')
        assert len(lstri) == nhead
        elmq = lstri[0].replace('<tr><td>', '')
        if '<sup>' in elmq:
            elm = elmq[:elmq.index('<sup>')]
            charge = elmq[elmq.index('<sup>')+len('<sup>'):]
            charge = charge.replace('</sup>', '')
            if charge == '+':
                charge = '1+'
        else:
            charge = ''
            elm = elmq
        typ = lstri[2][1:]
        fil = lstri[3][lstri[3].index('detail')+len('detail'):]
        fil = fil[:fil.index('.dat')+len('.dat')]
        lout.append([elm, charge, typ, fil])

    # Format output
    char = np.array(lout)
    col = ['Element', 'charge', 'type of data', 'full file name']
    arr = _getcharray(char, col=col,
                      sep='  ', line='-', just='l',
                      returnas=returnas, verb=verb)
    return arr


def search_online_by_wavelengthA(lambmin=None, lambmax=None, resolveby=None,
                                 element=None, charge=None,
                                 returnas=None, verb=None):
    """ Perform an online search by wavelength on https://open.adas.ac.uk

    Pass the min / max wavelength (in Angstrom) to the online wavelength search
    Prints the results (if verb=True)
    Optionally return the result as:
        - a char array (returnas = np.ndarray)
        - a formatted str (returnas = str)

    The result can be resolve by transition or by adas file
    Optionally filter by element to return only the results of one element

    example
    -------
        >>> import tofu as tf
        >>> tf.openadas2tofu.search_online_by_wavelengthA(3., 4., element='ar')
    """

    # Check input
    if returnas is None:
        returnas = False
    if verb is None:
        verb = True
    if lambmin is None:
        lambmin = ''
    if lambmax is None:
        lambmax = ''
    if resolveby is None:
        resolveby = 'transition'
    if resolveby not in ['transition', 'file']:
        msg = ("Arg resolveeby must be:\n"
               + "\t- 'transition': list all available transitions\n"
               + "\t- 'file': list all files containing relevant transitions")
        raise Exception(msg)
    if element is not None and not isinstance(element, str):
        msg = ("Arg element must be a str (e.g.: element='ar')\n"
               + "\t- provided: {}".format(element))
        raise Exception(msg)
    if charge is not None:
        if not isinstance(charge, int):
            msg = ("Arg charge must be a int!\n"
                   + "\t- provided: {}".format(charge))
            raise Exception(msg)
        charge = '0' if charge == 0 else '{}+'.format(charge)

    searchurl = '&'.join(['wave_min={}'.format(lambmin),
                          'wave_max={}'.format(lambmax),
                          'resolveby={}'.format(resolveby)])

    total_url = '{}{}&{}'.format(_URL_SEARCH_WAVL, searchurl, 'searching=1')
    resp = requests.get(total_url)

    # Extract response from html
    out = resp.text.split('\n')
    flag0 = '<table summary="Search by Wavelength Search Results">'
    flag1 = '</table></div></div></div>'
    ind0 = [ii for ii, vv in enumerate(out) if flag0 in vv]
    ind1 = [ii for ii, vv in enumerate(out) if flag1 in vv]
    if len(ind0) != 1 or len(ind1) != 1:
        msg = ("Format of html response seems to have changed!\n"
               + "Cannot find flags:\n"
               + "\t- {}\n".format(flag0)
               + "\t- {}\n".format(flag1)
               + "in requests.get({}).text".format(total_url))
        raise Exception(msg)
    out = out[ind0[0] + 1].split('</tr><tr><td>')
    nresults = len(out) - 1

    # Get columns
    col = [kk.replace('<tr><th>', '').replace('</th>', '').strip()
           for kk in out[0].split('</th><th>')]
    ncol = len(col)
    colex = ['Wavelength', 'Ion', 'Data Type', 'Transition', 'File Details']
    if col != colex:
        msg = ("Format of table columns in html seems to have changed!\n"
               + "\t- expected: {}\n".format(colex)
               + "\t- observed: {}".format(col))
        raise Exception(msg)

    lout = []
    for ii in range(0, nresults):
        lstri = out[ii+1].split('</td><td>')
        assert len(lstri) == ncol
        lamb = lstri[0].replace('&Aring;', '')
        elm, charg = lstri[1].replace('</sup>', '').split('<sup>')
        if charg == '+':
            charg = '1+'
        if element is not None and elm.lower() != element.lower():
            continue
        if charge is not None and charg != charge:
            continue
        typ = lstri[2].replace('</span>', '').split('>')[1]
        trans = lstri[3].replace('&nbsp;', ' ')
        trans = trans.replace('<sup>', '^{').replace('</sup>', '}')
        trans = trans.replace('<sub>', '_{').replace('</sub>', '}')
        trans = trans.replace('&rarr;', '->')
        fil = lstri[4][lstri[4].index('detail')+len('detail'):]
        fil = fil[:fil.index('.dat')+len('.dat')]
        lout.append([lamb, elm, charg, typ, trans, fil])

    # Format output
    char = np.array(lout)
    col = ['Wavelength', 'Element', 'Charge', 'Data Type',
           'Transition', 'Full file name']
    arr = _getcharray(char, col=col,
                      sep='  ', line='-', just='l',
                      returnas=returnas, verb=verb)
    return arr


# #############################################################################
#                          Download
# #############################################################################


def _check_exists(filename, update=None):

    # In case a small modification becomes necessary later
    target = filename
    path_local = _get_PATH_LOCAL()

    # Check whether the local .tofu repo exists, if not recommend tofu-custom
    if path_local is None:
        path = os.path.join(os.path.expanduser('~'), '.tofu', 'openadas2tofu')
        msg = ("You do not seem to have a local ./tofu repository\n"
               + "tofu uses that local repository to store all user-specific "
               + "data and downloads\n"
               + "In particular, openadas files are downloaded and saved in:\n"
               + "\t{}\n".format(path)
               + "  => to set-up your local .tofu repo, run in a terminal:\n"
               + "\ttofu-custom")
        raise Exception(msg)

    # Parse intermediate repos and create if necessary
    lrep = target.split('/')[1:-1]
    for ii in range(len(lrep)):
        repo = os.path.join(path_local, *lrep[:ii+1])
        if not os.path.isdir(repo):
            os.mkdir(repo)

    # Check if file already exists
    path = os.path.join(path_local, *lrep)
    pfe = os.path.join(path, target.split('/')[-1])
    if os.path.isfile(pfe):
        if update is False:
            msg = ("File already exists in your local repo:\n"
                   + "\t{}\n".format(pfe)
                   + "  => if you want to force download, use update=True"
                   + " (local file will be overwritten)")
            raise FileAlreayExistsException(msg)
        else:
            return True, pfe
    else:
        return False, pfe


def download(filename=None,
             update=None, verb=None, returnas=None):
    """ Download desired file from  https://open.adas.ac.uk

    All downloaded files are stored in your local tofu directory (~/.tofu/)

    example
    -------
        >>> import tofu as tf
        >>> filename = '/adf15/pec40][ar/pec40][ar_ls][ar16.dat'
        >>> tf.openadas2tofu.download(filename)
    """

    # ---------------------------
    # Check
    if verb is None:
        verb = True
    if update is None:
        update = False
    if returnas is None:
        returnas = False

    c0 = (not isinstance(filename, str)
          or filename[:4] != '/adf' or filename[-4:] != '.dat')
    if c0:
        msg = ("filename must be a str (full file name) of the form:\n"
               + "\t/adf.../.../....dat")
        raise Exception(msg)
    url = _URL_DOWNLOAD + filename

    exists, pfe = _check_exists(filename, update=update)
    if exists is True and verb is True:
        msg = ("File already exists, will be downloaded and overwritten:\n"
               + "\t{}".format(pfe))
        warnings.warn(msg)

    # ---------------------------
    # Download
    # Note the stream=True parameter below
    with requests.get(url, stream=True) as rr:
        rr.raise_for_status()
        with open(pfe, 'wb') as ff:
            for chunk in rr.iter_content(chunk_size=8192):
                # filter-out keep-alive new chunks
                if chunk:
                    ff.write(chunk)
                    # ff.flush()

    if verb is True:
        msg = ("file {} was copied to:\n".format(filename)
               + "\t{}".format(pfe))
        print(msg)
    if returnas is str:
        return pfe


def download_all(files=None, searchstr=None,
                 lambmin=None, lambmax=None, element=None,
                 include_partial=None, update=None, verb=None):
    """ Download all desired files from  https://open.adas.ac.uk

    The files to download can be provided either as:
        - a list of full openadas file names (files)
        - the result of an online freeform search
            (searchstr fed to search_online)
        - the result of an online search by wavelength
            (lambmin, lambmax and element fed to search_online_by_wavelengthA)

    All downloaded files are stored in your local tofu directory (~/.tofu/)

    example
    -------
        >>> import tofu as tf
        >>> tf.openadas2tofu.download_all(lambmin=3., lambmax=4., element='ar')
    """
    # Check
    if include_partial is None:
        include_partial = False
    if update is None:
        update = False
    if verb is None:
        verb = True
    lc = [files is not None,
          searchstr is not None,
          any([ss is not None for ss in [lambmin, lambmax, element]])]
    if np.sum(lc) != 1:
        msg = (
            "Please either searchstr xor (lambmin, lambmax, element)\n"
            + "\t- files: list of full file names to be downloaded\n"
            + "\t- searchstr: uses search_online()\n"
            + "\t- lambmin, lambmax, element: search_online_by_wavelengthA")
        raise Exception(msg)

    # Get list of files
    if lc[0]:
        if isinstance(files, str):
            files = [files]
        if not (isinstance(files, list)
                and all([isinstance(ss, str) for ss in files])):
            msg = "files must be a list of full openadas file names!"
            raise Exception(msg)
    elif lc[1]:
        arr = search_online(searchstr=searchstr,
                            include_partial=include_partial,
                            verb=False, returnas=np.ndarray)
        files = arr[:, -1]
    elif lc[2]:
        arr = search_online_by_wavelengthA(lambmin=lambmin,
                                           lambmax=lambmax,
                                           element=element, verb=False,
                                           returnas=np.ndarray)
        files = np.unique(arr[:, -1])

    # Download
    if verb is True:
        msg = "Downloading from {} into {}:".format(_URL, _get_PATH_LOCAL())
        print(msg)
    for ii in range(len(files)):
        try:
            exists, pfe = _check_exists(files[ii], update=update)
            pfe = download(filename=files[ii], update=update,
                           verb=False, returnas=str)
            if exists is True:
                msg = "\toverwritten:   \t{}".format(files[ii])
            else:
                msg = "\tdownloaded:    \t{}".format(files[ii])
        except FileAlreayExistsException:
            msg = "\talready exists:  {}".format(files[ii])
        except Exception as err:
            msg = (str(err)
                   + "\n\nCould not download file {}".format(files[ii]))
            raise err

        if verb is True:
            print(msg)


def clean_downloads():
    """ Delete all openadas files downloaded in your ~/.tofu/ directory """
    path_local = _get_PATH_LOCAL()
    if path_local is None:
        return
    lf = [ff for ff in os.listdir(path_local)
          if os.path.isfile(os.path.join(path_local, ff))]
    ld = [ff for ff in os.listdir(path_local)
          if os.path.isdir(os.path.join(path_local, ff))]
    for ff in lf:
        os.remove(ff)
    for dd in ld:
        shutil.rmtree(os.path.join(path_local, dd))










fs1D = (12,7)
fs2D = (12,7)

dmargin1D = dict(left=0.05, bottom=0.06, right=0.99, top=0.92,
                 wspace=0.3, hspace=0.5)
dmargin2D = dict(left=0.05, bottom=0.06, right=0.99, top=0.92,
                 wspace=0.3, hspace=0.8)
dmargin_combine = dict(left=0.05, bottom=0.06, right=0.99, top=0.92,
                       wspace=0.3, hspace=0.3)
